from io import StringIO
from unittest.mock import patch

from hera.main import main


def test_assemble_set_inc(capsys):
    main(["assemble", "--code", "--stdout", "test/assets/asm/set_inc.hera"])

    captured = capsys.readouterr()
    assert captured.err == ""

    with open("test/assets/asm/set_inc.hera.lcode") as f:
        assert captured.out == f.read()


def test_assemble_binary_op(capsys):
    main(["assemble", "--code", "--stdout", "test/assets/asm/binary_op.hera"])

    captured = capsys.readouterr()
    assert captured.err == ""

    with open("test/assets/asm/binary_op.hera.lcode") as f:
        assert captured.out == f.read()


def test_assemble_branch(capsys):
    main(["assemble", "--code", "--stdout", "test/assets/asm/branch.hera"])

    captured = capsys.readouterr()
    assert captured.err == ""

    with open("test/assets/asm/branch.hera.lcode") as f:
        assert captured.out == f.read()


def test_assemble_labelled_branch(capsys):
    main(["assemble", "--code", "--stdout", "test/assets/asm/branch_label.hera"])

    captured = capsys.readouterr()
    assert captured.err == ""

    with open("test/assets/asm/branch_label.hera.lcode") as f:
        assert captured.out == f.read()


def test_assemble_relative_branch(capsys):
    main(["assemble", "--code", "--stdout", "test/assets/asm/rel_branch.hera"])

    captured = capsys.readouterr()
    assert captured.err == ""

    with open("test/assets/asm/rel_branch.hera.lcode") as f:
        assert captured.out == f.read()


def test_assemble_flag(capsys):
    main(["assemble", "--code", "--stdout", "test/assets/asm/flag.hera"])

    captured = capsys.readouterr()
    assert captured.err == ""

    with open("test/assets/asm/flag.hera.lcode") as f:
        assert captured.out == f.read()


def test_assemble_shift(capsys):
    main(["assemble", "--code", "--stdout", "test/assets/asm/shift.hera"])

    captured = capsys.readouterr()
    assert captured.err == ""

    with open("test/assets/asm/shift.hera.lcode") as f:
        assert captured.out == f.read()


def test_assemble_misc(capsys):
    main(["assemble", "--code", "--stdout", "test/assets/asm/misc.hera"])

    captured = capsys.readouterr()
    assert captured.err == ""

    with open("test/assets/asm/misc.hera.lcode") as f:
        assert captured.out == f.read()


def test_assemble_debug_op(capsys):
    main(["assemble", "--code", "--stdout", "test/assets/asm/debug.hera"])

    captured = capsys.readouterr()
    assert captured.err == ""

    with open("test/assets/asm/debug.hera.lcode") as f:
        assert captured.out == f.read()


def test_assemble_data(capsys):
    main(["assemble", "--data", "--stdout", "test/assets/asm/data.hera"])

    captured = capsys.readouterr()
    assert captured.err == ""

    with open("test/assets/asm/data.hera.ldata") as f:
        assert captured.out == f.read()


def test_assemble_data_with_big_stack(capsys):
    main(["assemble", "--data", "--stdout", "--big-stack", "test/assets/asm/data.hera"])

    captured = capsys.readouterr()
    assert captured.err == ""

    with open("test/assets/asm/data_big_stack.hera.ldata") as f:
        assert captured.out == f.read()


def test_assemble_code_and_data_together(capsys):
    with patch("sys.stdin", StringIO("SET(R1, 42)")):
        main(["assemble", "--stdout", "-"])

    captured = capsys.readouterr()
    assert captured.err == ""
    assert (
        captured.out
        == """\

[DATA]
  49152*0
  c001

[CODE]
  e12a
  f100
"""
    )



"""All test programs in this module taken from the official HERA manual."""
from hera.data import DEFAULT_DATA_START
from hera.utils import to_u16
from .utils import execute_program_helper


def test_single_precision_arithmetic():
    # Figure 4.1, p. 25
    program = """\
SET(R2, 15)
SET(R3, 17)
SET(R4, -3)

CBON()

ADD(R1, R2, R3)
SETLO(Rt, 7)
MUL(R1, Rt, R1)
SETLO(Rt, 4)
MUL(Rt, Rt, R4)
ADD(R1, R1, Rt)

SUB(R5, R4, R3)
"""
    vm = execute_program_helper(program)

    assert vm.registers[1] == 212
    assert vm.registers[2] == 15
    assert vm.registers[3] == 17
    assert vm.registers[4] == to_u16(-3)
    assert vm.registers[5] == to_u16(-20)

    assert vm.flag_carry_block
    assert vm.flag_carry
    assert not vm.flag_overflow
    assert not vm.flag_zero
    assert vm.flag_sign


def test_double_precision_arithmetic():
    # Figure 4.2, p. 27
    program = """\
SET(R3, 17)
SET(R4, 4000)
SET(R5, 2)
SET(R6, 50000)
SET(R7, 500)
SET(R8, 3000)

CCBOFF()

COFF()
ADD(R2, R4, R6)
ADD(R1, R3, R5)
COFF()
SET(Rt, 16960)
ADD(R2, R2, Rt)
SETLO(Rt, 15)
ADD(R1, R1, Rt)

CON()
SUB(R8, R8, R2)
SUB(R7, R7, R1)
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == 35
    assert vm.registers[2] == 5424
    assert vm.registers[3] == 17
    assert vm.registers[4] == 4000
    assert vm.registers[5] == 2
    assert vm.registers[6] == 50000
    assert vm.registers[7] == 464
    assert vm.registers[8] == 63112

    assert not vm.flag_carry_block
    assert vm.flag_carry
    assert not vm.flag_overflow
    assert not vm.flag_zero
    assert not vm.flag_sign


def test_branching():
    # Figure 5.1, p. 30
    program = """\
CBON()
SETLO(R1, 0xB6)

CMP(R1, R0)
BGER(3)
NEG(R1, R1)
LSR(R1, R1)
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == 37


def test_global_variables():
    # Figure 6.1, p. 32
    program = """\
DLABEL(Variables)
DLABEL(X)
  INTEGER(12)
DLABEL(Y)
  DSKIP(1)
DLABEL(Z)
  INTEGER(4)

CBON()

SET(Rt, X)
LOAD(R1, 0, Rt)
INC(R1, 5)
SET(Rt, Y)
STORE(R1, 0, Rt)

SET(R1, Variables)
LOAD(R2, 0, R1)
LOAD(R3, 1, R1)
ADD(R3, R3, R3)
ADD(R2, R2, R3)
LOAD(R3, 2, R1)
SUB(R2, R2, R3)
STORE(R2, 0, R1)

HALT()
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == DEFAULT_DATA_START
    assert vm.registers[2] == 42
    assert vm.registers[3] == 4

    assert vm.memory[DEFAULT_DATA_START] == 42
    assert vm.memory[DEFAULT_DATA_START + 1] == 17
    assert vm.memory[DEFAULT_DATA_START + 2] == 4


def test_arrays():
    # Figure 6.2, p. 34
    program = """\
DLABEL(ArrayOfSevenPrimes)
  INTEGER(7)
  INTEGER(2) INTEGER(3) INTEGER(5) INTEGER(7) DSKIP(1) INTEGER(13) INTEGER(17)
DLABEL(ArrayOfSevenSquaredPrimes)
  DSKIP(8)

  CBON()

  SET(Rt, ArrayOfSevenPrimes)
  SETLO(R1, 5)
  ADD(Rt, Rt, R1)
  SETLO(R1, 11)
  STORE(R1, 0, Rt)

  SET(R1, ArrayOfSevenPrimes)
  SET(R2, ArrayOfSevenSquaredPrimes)
  LOAD(R3, 0, R1)
  STORE(R3, 0, R2)

LABEL(SquareNextOne)
  DEC(R3, 1)
BL(NoMoreSquares)
  INC(R1, 1)
  INC(R2, 1)
  LOAD(R4, 0, R1)
  MUL(R4, R4, R4)
  STORE(R4, 0, R2)
BR(SquareNextOne)

LABEL(NoMoreSquares)
HALT()
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == DEFAULT_DATA_START + 7
    assert vm.registers[2] == DEFAULT_DATA_START + 15
    assert vm.registers[3] == to_u16(-1)
    assert vm.registers[4] == 289

    assert vm.memory[DEFAULT_DATA_START] == 7
    assert vm.memory[DEFAULT_DATA_START + 1] == 2
    assert vm.memory[DEFAULT_DATA_START + 2] == 3
    assert vm.memory[DEFAULT_DATA_START + 3] == 5
    assert vm.memory[DEFAULT_DATA_START + 4] == 7
    assert vm.memory[DEFAULT_DATA_START + 5] == 11
    assert vm.memory[DEFAULT_DATA_START + 6] == 13
    assert vm.memory[DEFAULT_DATA_START + 7] == 17

    assert vm.memory[DEFAULT_DATA_START + 8] == 7
    assert vm.memory[DEFAULT_DATA_START + 9] == 4
    assert vm.memory[DEFAULT_DATA_START + 10] == 9
    assert vm.memory[DEFAULT_DATA_START + 11] == 25
    assert vm.memory[DEFAULT_DATA_START + 12] == 49
    assert vm.memory[DEFAULT_DATA_START + 13] == 121
    assert vm.memory[DEFAULT_DATA_START + 14] == 169
    assert vm.memory[DEFAULT_DATA_START + 15] == 289


def test_strings():
    # Figure 6.3, p. 35
    program = """\
DLABEL(The_string)
  LP_STRING("Is this an example? With three questions? Really?")

DLABEL(N_questions)
  INTEGER(0)

CBON()
SET(R1, 0)
SET(R2, The_string)
LOAD(R3, 0, R2)
INC(R2, 1)

LABEL(top_of_loop)
  LOAD(R4, 0, R2)
  SET(R5, 63)
  SUB(R0, R4, R5)
BNZ(not_a_question)
  INC(R1, 1)
LABEL(not_a_question)
  INC(R2, 1)
  DEC(R3, 1)
BNZ(top_of_loop)

SET(R2, N_questions)
STORE(R1, 0, R2)
HALT()
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == 3
    assert vm.registers[2] == 0xC033
    assert vm.registers[3] == 0
    assert vm.registers[4] == 63
    assert vm.registers[5] == 63

    for r in vm.registers[6:10]:
        assert r == 0

    s = "Is this an example? With three questions? Really?"
    assert vm.memory[DEFAULT_DATA_START] == len(s)
    for i in range(len(s)):
        assert vm.memory[DEFAULT_DATA_START + i + 1] == ord(s[i])

    assert vm.flag_carry_block
    assert vm.flag_carry
    assert not vm.flag_overflow
    assert vm.flag_zero
    assert not vm.flag_sign


def test_simple_function_calls():
    # Figure 7.4, p. 41
    program = """\
CBON()

SETLO(r1, 100)
SETLO(r2,  50)
CALL(FP_alt, updateR3)

SETLO(r1, 10)
SETLO(r2,  3)
CALL(FP_alt, updateR3)

HALT()

LABEL(updateR3)
  ADD(r1, r1, r1)
  ADD(r1, r1, r2)
  ADD(r3, r3, r1)
  RETURN(FP_alt, PC_ret)
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == 23
    assert vm.registers[2] == 3
    assert vm.registers[3] == 273


def test_call_to_library_function(capsys):
    # Figure 7.5, p. 43
    program = """\
#include <Tiger-stdlib-reg-data.hera>

DLABEL(INT_DIVIDE_AS_IN_PYTHON)  LP_STRING("//")
DLABEL(EQUALS_SIGN_WITH_SPACES)  LP_STRING(" = ")

CBON()
SET(R1, 210)
SET(R2, 5)

MOVE(FP_alt, SP)
CALL(FP_alt, printint)

MOVE(R4, R1)
MOVE(R5, R2)
SET(R1, INT_DIVIDE_AS_IN_PYTHON)
CALL(FP_alt, print)

MOVE(R1, R5)
CALL(FP_alt, printint)

SET(R1, EQUALS_SIGN_WITH_SPACES)
CALL(FP_alt, print)

MOVE(R1, R4)
MOVE(R2, R5)
CALL(FP_alt, div)
CALL(FP_alt, printint)

HALT()

#include <Tiger-stdlib-reg.hera>
"""
    execute_program_helper(program)

    captured = capsys.readouterr().out
    assert captured == "210//5 = 42"


def test_call_to_library_function_with_stack(capsys):
    # Figure 7.6, p. 44
    program = """\
#include <Tiger-stdlib-stack-data.hera>

DLABEL(INT_DIVIDE_AS_IN_PYTHON)  LP_STRING("//")
DLABEL(EQUALS_SIGN_WITH_SPACES)  LP_STRING(" = ")

CBON()
SET(R1, 210)
SET(R2, 5)

MOVE(FP_alt, SP)
INC(SP, 4)
STORE(R1, 3, FP_alt)
CALL(FP_alt, printint)
DEC(SP, 4)

MOVE(FP_alt, SP)
INC(SP, 4)
SET(Rt, INT_DIVIDE_AS_IN_PYTHON)
STORE(Rt, 3, FP_alt)
CALL(FP_alt, print)

STORE(R2, 3, FP_alt)
CALL(FP_alt, printint)

SET(Rt, EQUALS_SIGN_WITH_SPACES)
STORE(Rt, 3, FP_alt)
CALL(FP_alt, print)

INC(SP, 1)
STORE(R1, 3, FP_alt)
STORE(R2, 4, FP_alt)
CALL(FP_alt, div)
DEC(SP, 1)
CALL(FP_alt, printint)
DEC(SP, 4)

HALT()

#include <Tiger-stdlib-stack.hera>
"""
    execute_program_helper(program)

    captured = capsys.readouterr().out
    assert captured == "210//5 = 42"


def test_call_to_user_function(capsys):
    # Figures 7.7, 7.8, 7.9, pp. 45-47
    program = """\
#include <Tiger-stdlib-reg-data.hera>

CBON()

SET(R1, 5)

INC(SP, 1)
STORE(R1, 0, FP)
INC(R1, 5)
SET(R2, 2)
MOVE(FP_alt, SP)
CALL(FP_alt, foo)

LOAD(R2, 0, FP)
DEC(SP, 1)
SUB(R1, R1, R2)

CALL(FP_alt, printint)
HALT()

LABEL(foo)
  INC(SP, 3)
  STORE(PC_ret, 0, FP)
  STORE(FP_alt, 1, FP)
  STORE(R1, 2, FP)

  MOVE(R10, R1)
  MOVE(R11, R2)

  ADD(R1, R10, R11)
  SUB(R2, R11, R10)
  SETLO(R9, 75)
  ADD(R2, R2, R9)

  MOVE(FP_alt, SP)
  CALL(FP_alt, two_x_plus_y)
  LOAD(R2, 2, FP)
  MUL(R1, R1, R2)

  LOAD(PC_ret, 0, FP)
  LOAD(FP_alt, 1, FP)
  DEC(SP, 3)
  RETURN(FP_alt, PC_ret)

LABEL(two_x_plus_y)
  ADD(R1, R1, R1)
  ADD(R1, R1, R2)
  RETURN(FP_alt, PC_ret)

#include <Tiger-stdlib-reg.hera>
"""
    execute_program_helper(program)

    captured = capsys.readouterr().out
    assert captured == "905"


def test_call_to_user_function_with_stack(capsys):
    # Figures 7.11, 7.12, 7.14, pp. 48-51
    program = """\
#include <Tiger-stdlib-stack-data.hera>

CBON()

SET(R1, 5)

MOVE(FP_alt, SP)
INC(SP, 5)
MOVE(R2, R1)
INC(R2, 5)
STORE(R2, 3, FP_alt)
SET(R2, 2)
STORE(R2, 4, FP_alt)
CALL(FP_alt, foo)
LOAD(R2, 3, FP_alt)
DEC(SP, 5)

SUB(R2, R2, R1)
INC(SP, 4)
STORE(R2, 3, FP_alt)
CALL(FP_alt, printint)
DEC(SP, 4)

HALT()

LABEL(two_x_plus_y)
  INC(SP, 2)
  STORE(PC_ret, 0, FP)
  STORE(FP_alt, 1, FP)
  STORE(R1, 5, FP)
  STORE(R2, 6, FP)

  LOAD(R1, 3, FP)
  LOAD(R2, 4, FP)

  ADD(R1, R1, R1)
  ADD(R1, R1, R2)

  STORE(R1, 3, FP)

  LOAD(R2, 6, FP)
  LOAD(R1, 5, FP)
  LOAD(PC_ret, 0, FP)
  LOAD(FP_alt, 1, FP)
  DEC(SP, 2)
  RETURN(FP_alt, PC_ret)

LABEL(foo)
  INC(SP, 2)
  STORE(PC_ret, 0, FP)
  STORE(FP_alt, 1, FP)
  STORE(R1, 5, FP)
  STORE(R2, 6, FP)

  MOVE(FP_alt, SP)
  INC(SP, 5)

  LOAD(R1, 3, FP)
  LOAD(R2, 4, FP)
  ADD(Rt, R1, R2)
  STORE(Rt, 3, FP_alt)
  SUB(R2, R2, R1)
  SETLO(Rt, 75)
  ADD(R2, R2, Rt)
  STORE(R2, 4, FP_alt)

  CALL(FP_alt, two_x_plus_y)

  LOAD(R2, 3, FP_alt)
  DEC(SP, 5)
  MUL(R1, R2, R1)

  STORE(R1, 3, FP)

  LOAD(R2, 6, FP)
  LOAD(R1, 5, FP)
  LOAD(PC_ret, 0, FP)
  LOAD(FP_alt, 1, FP)
  DEC(SP, 2)
  RETURN(FP_alt, PC_ret)

#include <Tiger-stdlib-stack.hera>
"""
    execute_program_helper(program)

    captured = capsys.readouterr().out
    assert captured == "905"



"""
The HERA assembler.

The binary encodings of individual HERA operations are defined in op.py; this module
defines the imperative logic to convert a HERA program into machine code.

Author:  Ian Fisher (iafisher@fastmail.com)
Version: March 2019
"""
import textwrap

from .data import Program, Settings


def assemble(program: Program) -> "Tuple[List[bytes], List[bytes]]":
    """
    Assemble a program into machine code. The return value is (code, data), where
    `code` is a list of HERA operations encoded as bytes objects (two bytes per op), and
    `data` is a list of the initial contents of the data segment.
    """
    code = []
    data = []

    for op in program.code:
        assembled = op.assemble()
        if assembled is not None:
            code.append(assembled)

    for data_op in program.data:
        data.append(data_op.assemble())

    return (code, data)


def assemble_and_print(program: Program, settings: Settings) -> None:
    """
    Assemble a program into machine code, and print it to standard output.

    The format of the data segment is designed to be compatible with Logisim, the
    program used at Haverford to design microprocessors, and to mimic the behavior of
    Hassem, the assembler that hera-py replaces.
    """
    raw_code, raw_data = assemble(program)

    code = "\n".join(bytes_to_hex(b) for b in raw_code)

    raw_data_concat = b"".join(raw_data)
    datalist = []
    for i in range(0, len(raw_data_concat), 2):
        hi = raw_data_concat[i]
        lo = raw_data_concat[i + 1]
        datalist.append("{:x}".format((hi << 8) + lo))
    data = "\n".join(datalist)
    # I don't know what the significance of this cell is, but Hassem includes it.
    cell = (len(raw_data_concat) // 2) + settings.data_start
    data = "{:x}\n".format(cell) + data
    # Make sure to put zeroes up to the start of the data segment.
    # In Logisim, the syntax "x*0" means "Place x zeroes in memory."
    data = "{}*0\n".format(settings.data_start - 1) + data

    if settings.stdout:
        if settings.data:
            print(data)
        elif settings.code:
            print(code)
        else:
            print("[DATA]")
            print(textwrap.indent(data, "  "))
            print("[CODE]")
            print(textwrap.indent(code, "  "))
    else:
        if settings.path == "-":
            path = "stdin"
        else:
            path = settings.path

        with open(path + ".lcode", "w", encoding="ascii") as f:
            f.write(code)
            f.write("\n")

        with open(path + ".ldata", "w", encoding="ascii") as f:
            f.write(data)


def bytes_to_hex(b: bytes) -> str:
    """
    Implementation of the standard Python bytes.hex method, which is not available in
    Python 3.4.
    """
    try:
        return b.hex()
    except AttributeError:
        return "".join("{:0>2x}".format(c) for c in b)



"""
The lexer for HERA and the debugging mini-language.

Consumed by hera/parser.py and hera/debugger/miniparser.py.

Author:  Ian Fisher (iafisher@fastmail.com)
Version: February 2019
"""
import string

from hera.data import Location, Messages, Token
from hera.utils import is_register, PATH_STRING


class Lexer:
    """A lexer for HERA (and for the debugging mini-language)."""

    def __init__(self, text: str, *, path: "Optional[str]" = PATH_STRING) -> None:
        self.text = text
        self.file_lines = text.splitlines()
        if self.text.endswith("\n"):
            self.file_lines.append("")
        self.position = 0
        self.line = 1
        self.column = 1
        self.path = path
        self.messages = Messages()
        # Set the current token.
        self.next_token()

    def get_location(self) -> Location:
        """Return the current location of the lexer."""
        return Location(self.line, self.column, self.path, self.file_lines)

    def next_token(self) -> Token:
        """Advance forward one token, set self.tkn to it, and return it."""
        self.skip()

        if self.position >= len(self.text):
            self.set_token(Token.EOF, length=0)
        else:
            ch = self.text[self.position]
            if ch.isalpha() or ch == "_":
                length = self.read_symbol()
                if is_register(self.text[self.position : self.position + length]):
                    self.set_token(Token.REGISTER, length=length)
                else:
                    self.set_token(Token.SYMBOL, length=length)
            elif ch.isdigit():
                length = self.read_int()
                self.set_token(Token.INT, length=length)
            elif ch == '"':
                self.consume_str()
            elif ch == "'":
                self.consume_char()
            elif self.text[self.position :].startswith("#include"):
                self.set_token(Token.INCLUDE, length=len("#include"))
            elif ch == "<":
                self.consume_bracketed()
            elif ch == ":":
                self.position += 1
                length = self.read_symbol()
                self.set_token(Token.FMT, length=length)
            elif ch == "-":
                self.set_token(Token.MINUS)
            elif ch == "+":
                self.set_token(Token.PLUS)
            elif ch == "/":
                self.set_token(Token.SLASH)
            elif ch == "*":
                self.set_token(Token.ASTERISK)
            elif ch == "@":
                self.set_token(Token.AT)
            elif ch == "(":
                self.set_token(Token.LPAREN)
            elif ch == ")":
                self.set_token(Token.RPAREN)
            elif ch == "{":
                self.set_token(Token.LBRACE)
            elif ch == "}":
                self.set_token(Token.RBRACE)
            elif ch == ",":
                self.set_token(Token.COMMA)
            elif ch == ";":
                self.set_token(Token.SEMICOLON)
            else:
                self.set_token(Token.UNKNOWN)

        return self.tkn

    def read_int(self) -> int:
        """Read an integer starting at the current position, and return its length."""
        length = 1
        digits = {str(i) for i in range(10)}
        peek = self.peek_char()
        if self.text[self.position] == "0" and peek and peek in "boxBOX":
            length = 2
            if self.peek_char() in "xX":
                digits |= set(string.ascii_letters)

        while self.peek_char(length) in digits:
            length += 1

        return length

    def read_symbol(self) -> int:
        """Read a symbol starting at the current position, and return its length."""
        length = 1
        while True:
            ch = self.peek_char(length)
            if not (ch.isalpha() or ch.isdigit() or ch == "_"):
                break
            length += 1
        return length

    HEX_DIGITS = "0123456789abcdefABCDEF"

    def read_escape_char(self) -> "Tuple[str, int]":
        """
        Read an escape sequence (assuming `self.text[self.position] == "\\"`) and return
        a pair (value, length), where `value` is what the escape sequence resolves to
        and `length` is the number of characters read.
        """
        peek = self.peek_char()
        loc = self.get_location()
        loc = loc._replace(column=loc.column + 1)
        if peek == "":
            return ("", 0)
        elif peek == "x":
            # Hex escapes
            peek2 = self.peek_char(2)
            peek3 = self.peek_char(3)
            if peek2 in self.HEX_DIGITS and peek3 in self.HEX_DIGITS:
                ordv = int(peek2 + peek3, base=16)
                return (chr(ordv), 3)
            else:
                self.warn("invalid hex escape", loc)
                return ("x", 1)
        elif peek.isdigit():
            # Octal escapes
            length = 1
            while length <= 3:
                if not self.peek_char(length).isdigit():
                    break
                length += 1

            val = self.text[self.position + 1 : self.position + length]

            try:
                ordv = int(val, base=8)
            except ValueError:
                self.warn("invalid octal escape", loc)
                return (self.peek_char(), 1)
            else:
                return (chr(ordv), length - 1)
        else:
            escape = escape_char(peek)
            if len(escape) == 2:
                self.warn("unrecognized backslash escape", loc)
            return (escape, 1)

    def consume_bracketed(self) -> None:
        self.next_char()
        loc = self.get_location()
        start = self.position
        while self.position < len(self.text) and self.text[self.position] != ">":
            self.next_char()

        if self.position == len(self.text):
            self.tkn = Token(Token.ERROR, "unclosed bracketed expression", loc)
            return

        self.tkn = Token(Token.BRACKETED, self.text[start : self.position], loc)
        self.next_char()

    def consume_str(self) -> None:
        """
        Advance the lexer to one past the end of a string literal starting at the
        current position, and set self.tkn appropriately.
        """
        loc = self.get_location()
        self.next_char()
        s = self.consume_delimited('"')

        if self.position == len(self.text):
            self.tkn = Token(Token.ERROR, "unclosed string literal", loc)
            return

        self.next_char()
        self.tkn = Token(Token.STRING, s, loc)

    def consume_char(self) -> None:
        """
        Advance the lexer to one past the end of a character literal starting at the
        current position, and set self.tkn appropriately.
        """
        loc = self.get_location()
        self.next_char()
        s = self.consume_delimited("'")

        if self.position == len(self.text):
            self.tkn = Token(Token.ERROR, "unclosed character literal", loc)
            return

        self.next_char()

        if len(s) == 1:
            self.tkn = Token(Token.CHAR, s, loc)
        elif len(s) == 2 and s[0] == "\\":
            self.tkn = Token(Token.CHAR, s[1], loc)
        else:
            self.tkn = Token(Token.ERROR, "over-long character literal", loc)

    def consume_delimited(self, delimiter: str) -> str:
        """
        Advance the lexer to one past the end of an expression delimited by the
        character `delimiter`, and set self.tkn appropriately.

        Backslash escapes inside the expression are converted, and escaped delimiters
        are ignored.

        This method exists to capture the shared functionality of consuming string and
        character literals.
        """
        sbuilder = []
        while self.position < len(self.text) and self.text[self.position] != delimiter:
            if self.text[self.position] == "\\":
                value, length = self.read_escape_char()
                self.next_char()
                if length == 0:
                    # Length of zero indicates EOF.
                    break

                sbuilder.append(value)
                for _ in range(length):
                    self.next_char()
            else:
                sbuilder.append(self.text[self.position])
                self.next_char()

        return "".join(sbuilder)

    def skip(self) -> None:
        """Skip past whitespace and comments."""
        while True:
            # Skip whitespace.
            while self.position < len(self.text) and self.text[self.position].isspace():
                self.next_char()

            # Skip comments.
            if self.position < len(self.text) and self.text[self.position] == "/":
                if self.peek_char() == "/":
                    while self.position < len(self.text):
                        if self.text[self.position] == "\n":
                            break
                        self.next_char()
                elif self.peek_char() == "*":
                    # Move past the "/*" at the start of the comment.
                    self.next_char()
                    self.next_char()
                    while self.position < len(self.text):
                        if self.text[self.position] == "*" and self.peek_char() == "/":
                            break
                        self.next_char()
                    # Skip the "*/" at the end of the comment.
                    self.next_char()
                    self.next_char()
                else:
                    break
            else:
                break

    def next_char(self) -> None:
        """
        Advance the position in the text by one. Do not call this method if the current
        position is past the end of the text.
        """
        if self.text[self.position] == "\n":
            self.line += 1
            self.column = 1
        else:
            self.column += 1
        self.position += 1

    def peek_char(self, n=1) -> str:
        """
        Return the n'th character in the text past the current position. If past the
        end, return the empty string.
        """
        return (
            self.text[self.position + n] if self.position + n < len(self.text) else ""
        )

    def set_token(self, typ: str, *, length=1) -> None:
        """
        Set self.tkn to a Token object whose type is `typ` and whose value is the
        substring of the input of length `length` starting at the current position. The
        lexer's position is advanced to one past the end of the token.
        """
        loc = self.get_location()
        value = self.text[self.position : self.position + length]
        for _ in range(length):
            self.next_char()
        self.tkn = Token(typ, value, loc)

    def warn(self, msg: str, loc) -> None:
        self.messages.warn(msg, loc)


def escape_char(c):
    """
    Return the special character that `c` encodes.

        >>> escape_char("n")
        "\n"
    """
    if c == "n":
        return "\n"
    elif c == "t":
        return "\t"
    elif c == "\\":
        return "\\"
    elif c == '"':
        return '"'
    else:
        return "\\" + c



from hera.data import Token
from hera.op import match_bitvector, substitute_bitvector


def test_match_bitvector_with_literal_pattern():
    assert match_bitvector("0101 0101 0101 0101", 0b0101010101010101) == []
    assert match_bitvector("0101 0101 0101 0101", 0) is False


def test_match_bitvector_captures_one_register():
    m = match_bitvector("0000 1111 AAAA 0000", 0b0000111101010000)
    assert m == [Token(Token.REGISTER, 0b0101)]

    m = match_bitvector("0000 1111 AAAA 0000", 0b0000101101010010)
    assert m is False


def test_match_bitvector_captures_one_integer():
    m = match_bitvector("0000 1111 aaaa 0000", 0b0000111101010000)
    assert m == [Token(Token.INT, 0b0101)]


def test_match_bitvector_captures_multiple_args():
    m = match_bitvector("0101 aaaa BBBB cccc", 0b0101011010111111)
    assert m == [
        Token(Token.INT, 0b0110),
        Token(Token.REGISTER, 0b1011),
        Token(Token.INT, 0b1111),
    ]


def test_match_bitvector_captures_args_out_of_order():
    m = match_bitvector("0101 BBBB aaaa cccc", 0b0101011010111111)
    assert m == [
        Token(Token.INT, 0b1011),
        Token(Token.REGISTER, 0b0110),
        Token(Token.INT, 0b1111),
    ]


def test_match_bitvector_captures_split_arg():
    m = match_bitvector("dcba abcd dcba abcd", 0b0111000011100000)

    assert m == [
        Token(Token.INT, 0b1000),
        Token(Token.INT, 0b1010),
        Token(Token.INT, 0b1010),
        Token(Token.INT, 0b0010),
    ]


def test_substitute_bitvector_with_no_args():
    assert substitute_bitvector("0000 1111 0000 1111", []) == bytes([0b1111, 0b1111])


def test_substitute_bitvector_with_one_arg():
    assert substitute_bitvector("0000 aaaa 0000 1111", [7]) == bytes([7, 0b1111])


def test_substitute_bitvector_with_two_args():
    assert substitute_bitvector("0000 aaaa bbbb 1111", [7, 3]) == bytes([7, 63])


def test_substitute_bitvector_with_out_of_order_args():
    bv = substitute_bitvector("bb00 11bb a00a 1aa1", [0b1010, 0b0101])

    assert bv == bytes([0b01001101, 0b10001101])



import pytest

from hera.debugger.miniparser import (
    InfixNode,
    IntNode,
    MemoryNode,
    Lexer,
    MiniParser,
    PrefixNode,
    RegisterNode,
    SeqNode,
    SymbolNode,
)


def parse_helper(text, keep_seq=False):
    tree = MiniParser(Lexer(text)).parse()
    if not keep_seq and isinstance(tree, SeqNode):
        assert len(tree.seq) == 1
        return tree.seq[0]
    else:
        return tree


def test_parse_memory_expression():
    tree = parse_helper("@@rt")
    assert str(tree) == "@@R11"

    assert isinstance(tree, MemoryNode)
    assert isinstance(tree.address, MemoryNode)
    assert isinstance(tree.address.address, RegisterNode)
    assert tree.address.address.value == 11


def test_parse_memory_expression_with_integer():
    tree = parse_helper("@-0o12")
    assert str(tree) == "@-10"

    assert isinstance(tree, MemoryNode)
    assert isinstance(tree.address, PrefixNode)
    assert tree.address.op == "-"
    assert isinstance(tree.address.arg, IntNode)
    assert tree.address.arg.value == 0o12


def test_parse_memory_expression_with_symbol():
    tree = parse_helper("@add")
    assert str(tree) == "@add"

    assert isinstance(tree, MemoryNode)
    assert isinstance(tree.address, SymbolNode)
    assert tree.address.value == "add"


def test_parse_simple_arithmetic():
    tree = parse_helper("1 + 2 * 3")
    assert str(tree) == "1 + (2 * 3)"

    assert isinstance(tree, InfixNode)
    assert tree.op == "+"

    assert isinstance(tree.left, IntNode)
    assert tree.left.value == 1

    assert isinstance(tree.right, InfixNode)
    assert tree.right.op == "*"
    assert isinstance(tree.right.left, IntNode)
    assert isinstance(tree.right.right, IntNode)
    assert tree.right.left.value == 2
    assert tree.right.right.value == 3


def test_parse_grouped_arithmetic():
    tree = parse_helper("(1 + 2) * (3 + 4)")
    assert str(tree) == "(1 + 2) * (3 + 4)"

    assert isinstance(tree, InfixNode)
    assert tree.op == "*"

    assert isinstance(tree.left, InfixNode)
    assert tree.left.op == "+"
    assert isinstance(tree.left.left, IntNode)
    assert tree.left.left.value == 1
    assert isinstance(tree.left.right, IntNode)
    assert tree.left.right.value == 2

    assert isinstance(tree.right, InfixNode)
    assert tree.right.op == "+"
    assert isinstance(tree.right.left, IntNode)
    assert tree.right.left.value == 3
    assert isinstance(tree.right.right, IntNode)
    assert tree.right.right.value == 4


def test_parse_complicated_arithmetic():
    tree = parse_helper("@((1+2) /-0o123) + @@5 - r1")
    assert str(tree) == "(@((1 + 2) / -83) + @@5) - R1"

    assert isinstance(tree, InfixNode)
    assert tree.op == "-"

    assert isinstance(tree.left, InfixNode)
    assert tree.left.op == "+"
    assert isinstance(tree.left.left, MemoryNode)
    assert isinstance(tree.left.left.address, InfixNode)
    assert tree.left.left.address.op == "/"
    assert isinstance(tree.left.left.address.left, InfixNode)
    assert tree.left.left.address.left.op == "+"
    assert isinstance(tree.left.left.address.left.left, IntNode)
    assert isinstance(tree.left.left.address.left.right, IntNode)
    assert tree.left.left.address.left.left.value == 1
    assert tree.left.left.address.left.right.value == 2
    assert isinstance(tree.left.right, MemoryNode)
    assert isinstance(tree.left.right.address, MemoryNode)
    assert isinstance(tree.left.right.address.address, IntNode)
    assert tree.left.right.address.address.value == 5

    assert isinstance(tree.right, RegisterNode)
    assert tree.right.value == 1


def test_parse_tricky_precedence():
    tree = parse_helper("1*2+3")
    assert str(tree) == "(1 * 2) + 3"

    assert isinstance(tree, InfixNode)
    assert tree.op == "+"

    assert isinstance(tree.left, InfixNode)
    assert tree.left.op == "*"
    assert isinstance(tree.left.left, IntNode)
    assert tree.left.left.value == 1
    assert isinstance(tree.left.right, IntNode)
    assert tree.left.right.value == 2

    assert isinstance(tree.right, IntNode)
    assert tree.right.value == 3


def test_parse_expression_with_format_spec():
    tree = parse_helper(":y R1", keep_seq=True)
    assert str(tree) == ":y R1"

    assert isinstance(tree, SeqNode)
    assert tree.fmt == "y"
    assert len(tree.seq) == 1

    assert isinstance(tree.seq[0], RegisterNode)
    assert tree.seq[0].value == 1


def test_parse_sequence_of_expressions():
    tree = parse_helper("r1, @r2, 1 + 3", keep_seq=True)
    assert str(tree) == "R1, @R2, 1 + 3"

    assert isinstance(tree, SeqNode)
    assert tree.fmt == ""
    assert len(tree.seq) == 3

    assert isinstance(tree.seq[0], RegisterNode)
    assert tree.seq[0].value == 1

    assert isinstance(tree.seq[1], MemoryNode)
    assert isinstance(tree.seq[1].address, RegisterNode)
    assert tree.seq[1].address.value == 2

    assert isinstance(tree.seq[2], InfixNode)
    assert tree.seq[2].op == "+"
    assert isinstance(tree.seq[2].left, IntNode)
    assert isinstance(tree.seq[2].right, IntNode)
    assert tree.seq[2].left.value == 1
    assert tree.seq[2].right.value == 3


def test_parse_sequence_of_expressions_with_format_spec():
    tree = parse_helper(":xdc fp, sp", keep_seq=True)
    assert str(tree) == ":xdc R14, R15"

    assert isinstance(tree, SeqNode)
    assert tree.fmt == "xdc"
    assert len(tree.seq) == 2

    assert isinstance(tree.seq[0], RegisterNode)
    assert tree.seq[0].value == 14

    assert isinstance(tree.seq[1], RegisterNode)
    assert tree.seq[1].value == 15


def test_parse_with_trailing_input():
    with pytest.raises(SyntaxError):
        parse_helper("1+2 3")


def test_parse_with_invalid_octal_literal():
    with pytest.raises(SyntaxError):
        parse_helper("0o9")


def test_parse_with_unclosed_parenthesis():
    with pytest.raises(SyntaxError):
        parse_helper("(1 + 2")


def test_parse_with_unrecognized_character():
    with pytest.raises(SyntaxError):
        parse_helper("?1")


def test_parse_with_unexpected_token():
    with pytest.raises(SyntaxError):
        parse_helper(")1+2)")



from hera.data import Token
from hera.op import ADD, INC, LABEL, RTI, SET
from hera.parser import evaluate_ifdefs, parse


def valid(text, *, warnings=False):
    program, messages = parse(text)
    assert not messages.errors, messages.errors[0]
    if not warnings:
        assert not messages.warnings, messages.warnings[0]
    return program


def invalid(text):
    program, messages = parse(text)
    return messages


def test_parse_single_op():
    program = valid("SET(R1, 42)")

    assert program == [SET(R(1), INT(42))]


def test_parse_has_locations():
    program = valid("SET(  R1,\n42)")

    op = program[0]
    assert isinstance(op, SET)
    assert op.loc.line == 1
    assert op.loc.column == 1

    assert op.tokens[0].location.line == 1
    assert op.tokens[0].location.column == 7

    assert op.tokens[1].location.line == 2
    assert op.tokens[1].location.column == 1


def test_parse_op_with_no_args():
    program = valid("RTI()")

    assert program == [RTI()]


def test_parse_multiple_ops():
    program = valid("INC(R4, 5)\nADD(R1, R2, R3)")

    assert program == [INC(R(4), INT(5)), ADD(R(1), R(2), R(3))]


def test_parse_label_starting_with_register_name():
    program = valid("LABEL(R1_INIT)")

    assert program == [LABEL(Token.Sym("R1_INIT"))]


def test_parse_octal_number():
    program = valid("SETLO(R1, 0o123)")

    assert program[0].args[1] == 0o123


def test_parse_hex_number():
    program = valid("SETHI(R1, 0xaBc)")

    assert program[0].args[1] == 0xABC


def test_parse_binary_number():
    program = valid("INC(R7, 0b10101)")

    assert program[0].args[1] == 0b10101


def test_parse_empty_program():
    program = valid("")

    assert len(program) == 0


def test_parse_single_line_comment():
    program = valid("// A comment")

    assert len(program) == 0


def test_parse_single_line_comment_with_ops():
    program = valid("SETLO(R1, 1) // A comment\nSETHI(R1, 1)")

    assert len(program) == 2
    assert program[0].name == "SETLO"
    assert program[1].name == "SETHI"


def test_parse_cpp_boilerplate():
    program = valid(
        """\
#include <HERA.h>

void HERA_main() {
    SET(R1, 42)
}
    """,
        warnings=True,
    )

    assert len(program) == 1
    assert program[0].name == "SET"


def test_parse_cpp_boilerplate_weird_whitespace_and_spelling():
    program = valid("#include <HERA.h>\nvoid   HeRA_mAin( \t)\n {\n\n}", warnings=True)

    assert program == []


def test_parse_empty_cpp_boilerplate():
    program = valid(
        """\
#include <HERA.h>

void HERA_main() {
}
    """,
        warnings=True,
    )

    assert len(program) == 0


def test_parse_ops_with_semicolons():
    program = valid("SET(R1, 1); SET(R2, 2)")

    assert len(program) == 2
    assert program[0].name == "SET"
    assert program[1].name == "SET"


def test_evaluate_ifdefs_with_simple_ifdef():
    program = evaluate_ifdefs(
        """
#ifdef HERA_PY
  SET(R1, 42)
#endif
    """
    )

    assert program.strip() == "SET(R1, 42)"


def test_evaluate_ifdefs_with_no_ifdefs():
    original = "Just some text.\nWith no ifdefs."
    program = evaluate_ifdefs(original)
    assert program == original


def test_evaluate_ifdefs_with_ifdef_and_else():
    program = evaluate_ifdefs(
        """
#ifdef HERA_PY
  SET(R1, 42)
#else
  R1 = 42;
#endif
    """
    )

    assert program.strip() == "SET(R1, 42)"


def test_evaluate_ifdefs_with_ifdef_and_else_with_leading_whitespace():
    program = evaluate_ifdefs(
        """
  #ifdef HERA_PY
    SET(R1, 42)
  #else
    R1 = 42;
  #endif
    """
    )

    assert program.strip() == "SET(R1, 42)"


def test_evaluate_ifdefs_with_ifndef():
    program = evaluate_ifdefs(
        """
#ifndef HERA_PY
  R1 = 42;
#else
  SET(R1, 42)
#endif
    """
    )

    assert program.strip() == "SET(R1, 42)"


def test_evaluate_ifdefs_with_ifdef_of_unknown_token():
    program = evaluate_ifdefs(
        """
#ifndef HERA_CPP
  SET(R1, 42)
#else
  R1 = 42;
#endif
    """
    )

    assert program.strip() == "SET(R1, 42)"


def test_evaluate_ifdefs_with_complex_program():
    program = evaluate_ifdefs(
        """
// Leading text should be included.
#ifdef HERA_PY
  // 1: This should be included.
  #ifdef HERA_CPP
    // 1: This should not be included.
    #else
    // 2: This should be included.
    #endif
  // 3: This should be included.
  #endif
// 4: This should be included.
#ifndef HERA_PY
  // 2: This should not be included.
  #else
  // 5: This should be included.
  #ifdef WHATEVER
    // 3: This should not be included.
    #else
    // 6: This should be included.
  #endif
  // 7: This should be included.
#endif
// Trailing text should be included.
    """
    )

    assert (
        program.strip()
        == """\
// Leading text should be included.

  // 1: This should be included.

    // 2: This should be included.

  // 3: This should be included.

// 4: This should be included.

  // 5: This should be included.

    // 6: This should be included.

  // 7: This should be included.

// Trailing text should be included."""
    )


def SYM(s):
    return Token(Token.SYMBOL, s)


def INT(x):
    return Token(Token.INT, x)


def R(i):
    return Token(Token.REGISTER, i)



"""
Definitions of important data structures for hera-py.

Author:  Ian Fisher (iafisher@fastmail.com)
Version: February 2019
"""
from collections import namedtuple


# The default memory address of the first cell in the data segment.
DEFAULT_DATA_START = 0xC001

VOLUME_QUIET = "quiet"
VOLUME_NORMAL = "normal"
VOLUME_VERBOSE = "verbose"


class Settings:
    """Global settings of the interpreter."""

    def __init__(self, *, color=True, mode="", volume=VOLUME_NORMAL):
        # Are SWI and RTI operations allowed?
        self.allow_interrupts = False
        # Should the assembler print out code?
        self.code = False
        # Is color output enabled?
        self.color = color
        # Should the assembler print out data?
        self.data = False
        # Where is the start of the data segment?
        self.data_start = DEFAULT_DATA_START
        # How should the registers of the virtual machine be initialized?
        self.init = []
        # What is the program's mode (e.g., "debug", "assemble")?
        self.mode = mode
        # Are debugging operations allowed?
        self.no_debug_ops = False
        # Should the preprocessor obfuscate the given code?
        self.obfuscate = False
        # What path was the program invoked on?
        self.path = None
        # Should the assembler print to standard output?
        self.stdout = False
        # Should the interpreter quit after a certain number of operations have been
        # executed?
        self.throttle = False
        # Should warnings be issued for zero-prefixed octal numbers?
        self.warn_octal_on = True
        # Should warnings be issued for un-idiomatic use of the RETURN operation?
        self.warn_return_on = True
        # How loud should the program be?
        self.volume = volume
        # How many warnings have been issued? This isn't a setting, strictly speaking,
        # but it's convenient to attach it to this class.
        self.warning_count = 0


class Location(namedtuple("Location", ["line", "column", "path", "file_lines"])):
    """
    A class to represent a location in a file, for the use of warnings and error
    messages. The `file_lines` field is a pointer to a list of lines in the file, so
    that the actual line can be printed in the message.
    """

    def __repr__(self):
        if self.file_lines:
            return "Location(line={}, column={}, path={!r}, file_lines=[...])".format(
                self.line, self.column, self.path
            )
        else:
            return "Location(line={}, column={}, path={!r}, file_lines=[])".format(
                self.line, self.column, self.path
            )


# The data structure that represents a HERA program. `data` and `code` are each a list
# of AbstractOperations. `symbol_table` and `debug_info` are for the use of the
# debugger.
Program = namedtuple("Program", ["data", "code", "symbol_table", "debug_info"])
DebugInfo = namedtuple("DebugInfo", ["labels"])


class Token:
    # Possible values for the `type` field.
    INT = "TOKEN_INT"
    REGISTER = "TOKEN_REGISTER"
    SYMBOL = "TOKEN_SYMBOL"
    STRING = "TOKEN_STRING"
    BRACKETED = "TOKEN_BRACKETED"
    CHAR = "TOKEN_CHAR"
    MINUS = "TOKEN_MINUS"
    AT = "TOKEN_AT"
    ASTERISK = "TOKEN_ASTERISK"
    PLUS = "TOKEN_PLUS"
    SLASH = "TOKEN_SLASH"
    LPAREN = "TOKEN_LPAREN"
    RPAREN = "TOKEN_RPAREN"
    LBRACE = "TOKEN_LBRACE"
    RBRACE = "TOKEN_RBRACE"
    COMMA = "TOKEN_COMMA"
    SEMICOLON = "TOKEN_SEMICOLON"
    FMT = "TOKEN_FMT"
    INCLUDE = "TOKEN_INCLUDE"
    EOF = "TOKEN_EOF"
    ERROR = "TOKEN_ERROR"
    UNKNOWN = "TOKEN_UNKNOWN"

    def __init__(self, type_, value, location=None):
        self.type = type_
        # The value may be a string or an integer.
        self.value = value
        self.location = location

    @classmethod
    def R(cls, i, location=None):
        """Construct a register token."""
        return cls(cls.REGISTER, i, location)

    @classmethod
    def Int(cls, x, location=None):
        """Construct an integer token."""
        return cls(cls.INT, x, location)

    @classmethod
    def Sym(cls, s, location=None):
        """Construct a symbol token."""
        return cls(cls.SYMBOL, s, location)

    def __eq__(self, other):
        if isinstance(other, Token):
            return self.type == other.type and self.value == other.value
        else:
            # Convenience for comparing Tokens to strings and integers.
            return self.value == other

    def __repr__(self):
        return "Token({0.type!r}, {0.value!r}, loc={0.location!r})".format(self)


class Messages:
    """A class to represent a list of warning and error messages."""

    def __init__(self, msg=None, loc=None, *, warning=False):
        self.errors = []
        self.warnings = []
        if msg is not None:
            if warning:
                self.warnings.append((msg, loc))
            else:
                self.errors.append((msg, loc))

    def extend(self, messages: "Messages") -> "Messages":
        """Extend self with the warnings and errors in the other Messages object."""
        self.errors.extend(messages.errors)
        self.warnings.extend(messages.warnings)
        return self

    def err(self, msg: str, loc=None) -> None:
        """Record an error."""
        self.errors.append((msg, loc))

    def warn(self, msg: str, loc=None) -> None:
        """Record a warning."""
        self.warnings.append((msg, loc))


class HERAError(Exception):
    pass


class Constant(int):
    pass


class Label(int):
    pass


class DataLabel(int):
    pass



"""
The virtual HERA machine.

Author:  Ian Fisher (iafisher@fastmail.com)
Version: February 2019
"""
import copy
import sys

from .data import Program, Settings
from .utils import print_warning


class VirtualMachine:
    """
    An abstract representation of a HERA processor.

    This class defines the state of a HERA processor and some utility functions for
    manipulating it, but the HERA language itself is defined in `hera/op.py`.
    """

    def __init__(self, settings=Settings()) -> None:
        self.settings = settings
        self.reset()

    def reset(self) -> None:
        """Reset the machine to its initial state."""
        # Sixteen 16-bit registers. The virtual machine stores integers in their
        # unsigned representation, so the values of self.registers will always be
        # non-negative, although values above 2**15 implicitly represent negative
        # integers under a signed interpretation.
        self.registers = [0] * 16
        # 16-bit program counter
        self.pc = 0
        # Current memory cell for data instructions
        self.dc = self.settings.data_start
        # Status/control flags
        self.flag_sign = False
        self.flag_zero = False
        self.flag_overflow = False
        self.flag_carry = False
        self.flag_carry_block = False
        # A memory array of 16-bit words. The HERA specification requires 2**16 words
        # to be addressable, but we start off with a considerably smaller array and
        # expand it as necessary, to keep the start-up time fast.
        self.memory = [0] * (2 ** 4)
        # Used by some Tiger standard library functions for rudimentary IO.
        self.input_buffer = ""
        self.input_pos = 0
        # Stack of (call_address, return_address) pairs for CALL/RETURN instructions.
        # Used for warning messages and debugging.
        self.expected_returns = []  # type: List[Tuple[int, int]]
        # Special flag set by the HALT operation.
        self.halted = False
        # Location object for the current operation
        self.location = None
        # Number of operations that have been executed - used for throttling.
        self.op_count = 0
        # Have warnings been issued for use of SWI and RTI instructions?
        self.warned_for_SWI = False
        self.warned_for_RTI = False
        self.warned_for_overflow = False
        self.warning_count = 0

        # Initialize registers according to --init flag.
        for dest, val in self.settings.init:
            self.registers[dest] = val

    def copy(self) -> "VirtualMachine":
        """Return a copy of the virtual machine."""
        ret = copy.copy(self)
        ret.registers = self.registers.copy()
        ret.memory = self.memory.copy()
        return ret

    def run(self, program: Program) -> None:
        """Execute a program, resetting the machine's state beforehand."""
        self.reset()

        for data_op in program.data:
            data_op.execute(self)

        # This loop is performance-critical, so instead of having a single loop that
        # always does the throttle-checking, we check beforehand if throttling is turned
        # on, to avoid the performance penalty in the (normal) case where throttling is
        # off.
        if self.settings.throttle is False:
            while not self.halted and self.pc < len(program.code):
                op = program.code[self.pc]
                self.location = op.loc
                op.execute(self)
        else:
            while (
                not self.halted
                and self.pc < len(program.code)
                and self.op_count < self.settings.throttle
            ):
                op = program.code[self.pc]
                self.location = op.loc
                op.execute(self)
                self.op_count += 1

    def load_register(self, index: int) -> int:
        """Get the contents of the register with the given index."""
        return self.registers[index]

    def store_register(self, index: int, value: int) -> None:
        """Store the value in the target register."""
        if index != 0:
            self.registers[index] = value
            if index == 15 and value >= self.settings.data_start:
                if not self.warned_for_overflow:
                    self.warn(
                        "stack has overflowed into data segment", loc=self.location
                    )
                    self.warned_for_overflow = True

    def set_zero_and_sign(self, value: int) -> None:
        """Set the zero and sign flags based on the value."""
        self.flag_zero = value == 0
        self.flag_sign = bool(value & 0x8000)

    def load_memory(self, address: int) -> int:
        """Get the value at the given memory address."""
        if address >= len(self.memory):
            return 0
        else:
            return self.memory[address]

    def store_memory(self, address: int, value: int) -> None:
        """Store a value to a location in memory."""
        # Extend the size of the memory array if necessary.
        if address >= len(self.memory):
            self.memory.extend([0] * (address - len(self.memory) + 1))
        self.memory[address] = value

    def readline(self) -> None:
        """Read a line from standard input."""
        self.input_buffer = sys.stdin.readline().rstrip("\n")

    def warn(self, msg: str, loc) -> None:
        """Print a warning message."""
        print_warning(self.settings, msg, loc=loc)
        self.warning_count += 1



import pytest
from .utils import helper

from hera.vm import VirtualMachine


@pytest.fixture
def vm():
    return VirtualMachine()


def test_SAVEF_with_sign(vm):
    vm.flag_sign = True

    helper(vm, "SAVEF(R5)")

    assert vm.registers[5] == 1
    assert vm.flag_sign


def test_SAVEF_with_zero(vm):
    vm.flag_zero = True

    helper(vm, "SAVEF(R5)")

    assert vm.registers[5] == 0b10
    assert vm.flag_zero


def test_SAVEF_with_overflow(vm):
    vm.flag_overflow = True

    helper(vm, "SAVEF(R5)")

    assert vm.registers[5] == 0b100
    assert vm.flag_overflow


def test_SAVEF_with_carry(vm):
    vm.flag_carry = True

    helper(vm, "SAVEF(R5)")

    assert vm.registers[5] == 0b1000
    assert vm.flag_carry


def test_SAVEF_with_carry_block(vm):
    vm.flag_carry_block = True

    helper(vm, "SAVEF(R5)")

    assert vm.registers[5] == 0b10000


def test_SAVEF_with_several_flags(vm):
    vm.flag_sign = True
    vm.flag_overflow = True
    vm.flag_carry = True

    helper(vm, "SAVEF(R5)")

    assert vm.registers[5] == 0b1101
    assert vm.flag_sign
    assert vm.flag_overflow
    assert vm.flag_carry


def test_SAVEF_with_all_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True

    helper(vm, "SAVEF(R5)")

    assert vm.registers[5] == 0b11111
    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_SAVEF_with_no_flags(vm):
    helper(vm, "SAVEF(R5)")

    assert vm.registers[5] == 0


def test_SAVEF_overwrites_high_bits(vm):
    vm.registers[5] = 17500

    helper(vm, "SAVEF(R5)")

    assert vm.registers[5] == 0


def test_SAVEF_increments_pc(vm):
    helper(vm, "SAVEF(R5)")

    assert vm.pc == 1


def test_SAVEF_does_not_affect_R0(vm):
    vm.flag_carry = True

    helper(vm, "SAVEF(R0)")

    assert vm.registers[0] == 0


def test_RSTRF_with_sign(vm):
    vm.registers[5] = 1

    helper(vm, "RSTRF(R5)")

    assert vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_RSTRF_with_zero(vm):
    vm.registers[5] = 0b10

    helper(vm, "RSTRF(R5)")

    assert not vm.flag_sign
    assert vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_RSTRF_with_overflow(vm):
    vm.registers[5] = 0b100

    helper(vm, "RSTRF(R5)")

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_RSTRF_with_carry(vm):
    vm.registers[5] = 0b1000

    helper(vm, "RSTRF(R5)")

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert not vm.flag_carry_block


def test_RSTRF_with_carry_block(vm):
    vm.registers[5] = 0b10000

    helper(vm, "RSTRF(R5)")

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert vm.flag_carry_block


def test_RSTRF_with_several_flags(vm):
    vm.registers[5] = 0b1101

    helper(vm, "RSTRF(R5)")

    assert vm.flag_sign
    assert not vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert not vm.flag_carry_block


def test_RSTRF_with_all_flags(vm):
    vm.registers[5] = 0b11111

    helper(vm, "RSTRF(R5)")

    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_RSTRF_with_no_flags(vm):
    vm.flag_sign = True
    vm.flag_zero = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True

    helper(vm, "RSTRF(R5)")

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_RSTRF_increments_pc(vm):
    helper(vm, "RSTRF(R5)")

    assert vm.pc == 1


def test_FON_with_sign(vm):
    helper(vm, "FON(1)")

    assert vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_FON_with_zero(vm):
    helper(vm, "FON(0b10)")

    assert not vm.flag_sign
    assert vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_FON_with_overflow(vm):
    helper(vm, "FON(0b100)")

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_FON_with_carry(vm):
    helper(vm, "FON(0b1000)")

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert not vm.flag_carry_block


def test_FON_with_carry_block(vm):
    helper(vm, "FON(0b10000)")

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert vm.flag_carry_block


def test_FON_with_multiple_flags(vm):
    helper(vm, "FON(0b10101)")

    assert vm.flag_sign
    assert not vm.flag_zero
    assert vm.flag_overflow
    assert not vm.flag_carry
    assert vm.flag_carry_block


def test_FON_with_no_flags(vm):
    helper(vm, "FON(0)")

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_FON_does_not_overwrite_flags(vm):
    vm.flag_carry_block = True

    helper(vm, "FON(1)")

    assert vm.flag_sign
    assert vm.flag_carry_block


def test_FON_increments_pc(vm):
    helper(vm, "FON(0)")

    assert vm.pc == 1


def test_FOFF_with_sign(vm):
    vm.flag_zero = True
    vm.flag_sign = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True

    helper(vm, "FOFF(1)")

    assert not vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_FOFF_with_zero(vm):
    vm.flag_zero = True
    vm.flag_sign = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True

    helper(vm, "FOFF(0b10)")

    assert vm.flag_sign
    assert not vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_FOFF_with_overflow(vm):
    vm.flag_zero = True
    vm.flag_sign = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True

    helper(vm, "FOFF(0b100)")

    assert vm.flag_sign
    assert vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_FOFF_with_carry(vm):
    vm.flag_zero = True
    vm.flag_sign = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True

    helper(vm, "FOFF(0b1000)")

    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert not vm.flag_carry
    assert vm.flag_carry_block


def test_FOFF_with_carry_block(vm):
    vm.flag_zero = True
    vm.flag_sign = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True

    helper(vm, "FOFF(0b10000)")

    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert not vm.flag_carry_block


def test_FOFF_with_multiple_flags(vm):
    vm.flag_zero = True
    vm.flag_sign = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True

    helper(vm, "FOFF(0b10101)")

    assert not vm.flag_sign
    assert vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert not vm.flag_carry_block


def test_FOFF_with_no_flags(vm):
    vm.flag_zero = True
    vm.flag_sign = True
    vm.flag_overflow = True
    vm.flag_carry = True
    vm.flag_carry_block = True

    helper(vm, "FOFF(0)")

    assert vm.flag_sign
    assert vm.flag_zero
    assert vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block


def test_FOFF_increments_pc(vm):
    helper(vm, "FOFF(0)")

    assert vm.pc == 1


def test_FSET5_with_sign(vm):
    helper(vm, "FSET5(1)")

    assert vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_FSET5_with_zero(vm):
    helper(vm, "FSET5(0b10)")

    assert not vm.flag_sign
    assert vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_FSET5_with_overflow(vm):
    helper(vm, "FSET5(0b100)")

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_FSET5_with_carry(vm):
    helper(vm, "FSET5(0b1000)")

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert not vm.flag_carry_block


def test_FSET5_with_carry_block(vm):
    helper(vm, "FSET5(0b10000)")

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert vm.flag_carry_block


def test_FSET5_with_multiple_flags(vm):
    helper(vm, "FSET5(0b10101)")

    assert vm.flag_sign
    assert not vm.flag_zero
    assert vm.flag_overflow
    assert not vm.flag_carry
    assert vm.flag_carry_block


def test_FSET5_with_no_flags(vm):
    helper(vm, "FSET5(0)")

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_FSET5_does_overwrite_flags(vm):
    vm.flag_zero = True
    vm.flag_carry_block = True

    helper(vm, "FSET5(1)")

    assert vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_carry_block


def test_FSET5_increments_pc(vm):
    helper(vm, "FSET5(0)")

    assert vm.pc == 1


def test_FSET4_with_sign(vm):
    helper(vm, "FSET4(1)")

    assert vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_FSET4_with_zero(vm):
    helper(vm, "FSET4(0b10)")

    assert not vm.flag_sign
    assert vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_FSET4_with_overflow(vm):
    helper(vm, "FSET4(0b100)")

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_FSET4_with_carry(vm):
    helper(vm, "FSET4(0b1000)")

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert not vm.flag_carry_block


def test_FSET4_with_multiple_flags(vm):
    helper(vm, "FSET4(0b0101)")

    assert vm.flag_sign
    assert not vm.flag_zero
    assert vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_FSET4_with_no_flags(vm):
    helper(vm, "FSET4(0)")

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_FSET4_does_overwrite_flags(vm):
    vm.flag_zero = True
    vm.flag_overflow = True

    helper(vm, "FSET4(1)")

    assert vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow


def test_FSET4_increments_pc(vm):
    helper(vm, "FSET4(0)")

    assert vm.pc == 1



"""
The interactive shell interface to the HERA debugger.

Author:  Ian Fisher (iafisher@fastmail.com)
Version: March 2019
"""
import copy
import functools
import textwrap

from hera.assembler import assemble_and_print
from hera.data import DataLabel, HERAError, Label, Location, Program, Settings
from hera.loader import load_program
from hera.op import LABEL, OPCODE, Branch, DataOperation, disassemble, name_to_class
from hera.parser import parse
from hera.utils import format_int, out_of_range, pad

from . import miniparser
from .debugger import Debugger
from .miniparser import (
    AbstractNode,
    InfixNode,
    IntNode,
    MemoryNode,
    PrefixNode,
    RegisterNode,
    SymbolNode,
)


def debug(program: Program, settings: Settings) -> None:
    """Begin an interactive debugging session with the given program."""
    debugger = Debugger(program, settings)
    Shell(debugger, settings).loop()


def mutates(handler):
    """
    Decorator for command handlers in the Shell class that mutate the state of the
    debugger.
    """

    @functools.wraps(handler)
    def inner(self, *args, **kwargs):
        self.debugger.save()
        name = handler.__name__[len("handle_") :]
        self.command_history.append(name)
        return handler(self, *args, **kwargs)

    return inner


class Shell:
    """
    A class for the interactive command-line interface to the debugger.

    The bulk of this class consists of the command handler methods, one for each
    debugging command. A command `foo` should have a corresponding command handler named
    `handle_foo`. If `foo` changes the state of the debugger or shell, then it should
    be decorated with `@mutates`. The name of the command should be added either to the
    CAN_BE_ABBREVIATED or CANNOT_BE_ABBREVIATED class-level lists, as appropriate, and
    if the command needs to accept the entire argument list as a string rather than a
    list of strings (e.g., because whitespace is significant), it should be added to the
    TAKES_ARGSTR list. The docstring of the command handler is displayed to the user by
    the help command (e.g., `help foo`). See one of the existing command handlers for an
    example of the required docstring format.
    """

    def __init__(self, debugger: Debugger, settings: Settings) -> None:
        self.debugger = debugger
        self.settings = settings
        self.command_history = []  # type: List[str]

    def loop(self) -> None:
        """Run the debug loop."""
        if self.debugger.empty():
            print("Cannot debug an empty program.")
            return

        print("Welcome to the HERA debugger.")
        print()
        print("Enter 'help' to see a list of valid commands.")
        print("Enter 'quit' or press Ctrl+D to exit.")
        print()
        print()

        self.print_current_op()

        previous = None
        while True:
            try:
                response = input(">>> ").strip()
            except (EOFError, KeyboardInterrupt):
                print()
                break

            if not response:
                if previous:
                    response = previous
                    print("(executing previous command: {})".format(previous))
                    print()
                else:
                    continue

            should_continue = self.handle_command(response)
            if should_continue is False:
                break

            previous = response

    def handle_command(self, response: str) -> bool:
        """
        Parse the command and execute it. Return False if the loop should exit, and
        True otherwise.
        """
        try:
            cmd, argstr = response.split(maxsplit=1)
        except ValueError:
            cmd = response
            argstr = ""

        try:
            fullcmd = self.expand_command(cmd)
        except HERAError:
            if "=" in response:
                self.handle_assign(response.split("=", maxsplit=1))
            else:
                print("{} is not a recognized command.".format(cmd))

            return True
        else:
            if fullcmd == "quit":
                return False

            handler = getattr(self, "handle_" + fullcmd)
            if fullcmd in self.TAKES_ARGSTR:
                handler(argstr)
            else:
                handler(argstr.split())

            return True

    # Commands which may be abbreviated with a prefix. Order determines precedence when
    # multiple commands share a prefix.
    CAN_BE_ABBREVIATED = (
        # Multiple lists to prevent reformatting.
        ["assign", "break", "continue", "clear", "execute", "goto"]
        + ["help", "info", "list", "next", "print", "quit", "step", "undo"]
    )

    # Commands that require the whole command to be spelled out.
    CANNOT_BE_ABBREVIATED = ["asm", "dis", "doc", "ll", "off", "on", "restart"]

    # Commands that do not take a whitespace-separated list of argument.
    TAKES_ARGSTR = ["asm", "execute", "print"]

    def expand_command(self, cmd: str) -> str:
        """
        Expand an abbreviated command name into its full name, or raise HERAError if the
        abbreviation is not recognized.
        """
        cmd = cmd.lower()
        for full in self.CAN_BE_ABBREVIATED:
            if full.startswith(cmd):
                return full

        for full in self.CANNOT_BE_ABBREVIATED:
            if full == cmd:
                return full

        raise HERAError

    def handle_asm(self, argstr: str) -> None:
        """
        asm <op>
          Assemble the HERA operation into machine code
        """
        if not argstr.strip():
            print("asm takes one argument.")
            return

        # Check if the program is all code, all data, or a mix.
        any_code = False
        any_data = False
        for op in parse(argstr)[0]:
            if isinstance(op, DataOperation):
                any_data = True
            else:
                any_code = True

        try:
            program = load_program(argstr, self.settings)
        except SystemExit:
            return

        code_flag = any_code and not any_data
        data_flag = any_data and not any_code
        settings = self.asm_settings(code=code_flag, data=data_flag)
        assemble_and_print(program, settings)

    @mutates
    def handle_assign(self, args: "List[str]") -> None:
        """
        assign <x> <y>
          Assign the value of y to x. x may be a register, a memory location, or the
          program counter. y may be a register, a memory location, the program counter,
          a symbol, or an integer.

        <x> = <y>
          Alias for "assign <x> <y>", with the additional advantage that <x> and <y>
          may contain spaces.

          Examples:
            Assign to a register:          R1 = 42
            Assign to a memory location:   @(1000) = R4
            Assign a label to a register:  R1 = some_label
            Arithmetic:                    R7 = R5 * 10
        """
        if len(args) != 2:
            print("assign takes two arguments.")
            return

        try:
            ltree = miniparser.parse(args[0])
            rtree = miniparser.parse(args[1])
        except SyntaxError as e:
            msg = str(e)
            if msg:
                print("Parse error: " + msg + ".")
            else:
                print("Parse error.")
            return

        if len(ltree.seq) > 1:
            print("Parse error: cannot assign to sequence.")
            return

        if len(rtree.seq) > 1:
            print("Parse error: cannot assign sequence value.")
            return

        ltree = ltree.seq[0]
        rtree = rtree.seq[0]

        vm = self.debugger.vm
        try:
            rhs = self.evaluate_node(rtree)
            if isinstance(ltree, RegisterNode):
                vm.store_register(ltree.value, rhs)
            elif isinstance(ltree, MemoryNode):
                address = self.evaluate_node(ltree.address)
                vm.store_memory(address, rhs)
            elif isinstance(ltree, SymbolNode):
                if ltree.value == "pc":
                    vm.pc = rhs
                else:
                    print("Eval error: cannot assign to symbol.")
            elif isinstance(ltree, (InfixNode, PrefixNode)):
                print("Eval error: cannot assign to arithmetic expression.")
            else:
                raise RuntimeError(
                    "unknown node type {}".format(ltree.__class__.__name__)
                )
        except HERAError as e:
            print("Eval error: " + str(e) + ".")

    @mutates
    def handle_break(self, args: "List[str]") -> None:
        """
        break
          Print all current breakpoints.

        break <n>
          Set a breakpoint at the given line number in the file that the debugger was
          opened on.

        break <path>:<n>
          Set a breakpoint at the given line number in the given file.

        break <label>
          Set a breakpoint at the given label.
          
        break .
          Set a breakpoint at the current instruction.
        """
        if len(args) > 1:
            print("break takes zero or one arguments.")
            return

        if len(args) == 0:
            breakpoints = self.debugger.get_breakpoints()
            if breakpoints:
                for brk in breakpoints.values():
                    print(brk)
            else:
                print("No breakpoints set.")
        else:
            try:
                b = self.debugger.location_to_instruction_number(args[0])
            except ValueError as e:
                print("Error:", e)
            else:
                self.debugger.set_breakpoint(b)
                loc = self.debugger.op(b).loc
                print("Breakpoint set in file {0.path}, line {0.line}.".format(loc))

    @mutates
    def handle_clear(self, args: "List[str]") -> None:
        """
        clear <location>
          Clear a breakpoint at the given line. Location formats accepted are the same
          as the break command.

        clear *
          Clear all breakpoints.
        """
        if len(args) == 0:
            print("clear takes one or more arguments.")
            return

        if any(arg == "*" for arg in args):
            self.debugger.breakpoints.clear()
            print("Cleared all breakpoints.")
        else:
            for arg in args:
                try:
                    b = self.debugger.location_to_instruction_number(arg)
                except ValueError as e:
                    print("Error:", e)
                else:
                    if b in self.debugger.breakpoints:
                        loc = self.debugger.op(b).loc
                        del self.debugger.breakpoints[b]
                        msg = "Cleared breakpoint in file {0.path}, line {0.line}."
                        print(msg.format(loc))
                    else:
                        print("No breakpoint at that location.")

    @mutates
    def handle_continue(self, args: "List[str]") -> None:
        """
        continue
          Execute the program until a breakpoint is encountered or the program
          terminates.
        """
        if len(args) != 0:
            print("continue takes no arguments.")
            return

        self.debugger.next(step=True)
        while not self.debugger.finished() and not self.debugger.at_breakpoint():
            self.debugger.next(step=True)

        self.print_current_op()

    def handle_dis(self, args: "List[str]") -> None:
        """
        dis <n>
          Interpret the 16-bit integer as a HERA machine instruction, and disassemble
          it into its assembly-language mnemonic.

        dis <n1> <n2>...
          Disassemble multiple integers.

        dis
          If the current instruction is an OPCODE, disassemble its contents.
        """
        if len(args) > 0:
            intargs = []
            for a in args:
                try:
                    v = int(a, base=0)
                except ValueError:
                    print("Could not parse argument `{}` to dis.".format(a))
                    return
                else:
                    intargs.append(v)

            for v in intargs:
                try:
                    print(disassemble(v))
                except HERAError as e:
                    print("Error:", e)
        else:
            if not self.debugger.finished() and isinstance(self.debugger.op(), OPCODE):
                op = self.debugger.op()
                print(disassemble(op.args[0]))
            else:
                print("Current operation is not an OPCODE.")

    _doc_branch_message = """
      HERA supports two kinds of branching instructions: register branching and
      relative branching.

      Register branching
        Register branching instructions take a label argument and jump to the label if
        the instruction's condition is met. If the condition is not met, execution
        continues to the next instruction.

        Register branching instructions may also take a register argument instead of a
        label (hence the name), in which case instead of jumping to a label they jump to
        the n'th instruction where n is the contents of the register. In most cases, you
        want to use a label rather than a register.

      Relative branching
        For every register branching instruction, there is a counterpart relative
        branching instruction, whose name is the same except with an extra 'R' at the
        end. For example, the register branching instruction BULE has a relative
        branching counterpart called BULER.

        Relative branching instructions take an integer argument instead of a label or
        a register, and if their condition is met they jump that many instructions
        forward or backwards (depending on if the number is positive or negative).

        Relative branching instructions can also take a label argument, in which case
        they do the same thing as their corresponding register branching instruction.

        In most cases, HERA programmers should use register branching instructions with
        labels, and avoid using relative branching instructions at all.
    """

    def handle_doc(self, args: "List[str]") -> None:
        """
        doc <opname>...
          For each operation, print a message detailing its use and behavior.

        doc
          Same as above, except that the documentation for the current operation is
          printed.
        """
        if not args:
            args = [self.debugger.op().name]

        for arg in args:
            arg = arg.upper()
            if arg == "BRANCH":
                docstring = self._doc_branch_message
            else:
                try:
                    op = name_to_class[arg]
                except KeyError:
                    print("{} is not a HERA operation.".format(arg))
                    continue
                else:
                    docstring = op.__doc__

            if not docstring:
                print("{} has no documentation.".format(arg))
            else:
                docstring = textwrap.dedent(docstring)
                print(docstring)

    @mutates
    def handle_execute(self, argstr: str) -> None:
        """
        execute <op>
          Execute a HERA operation. The operation must not be a data statement or a
          branch. The operation may affect registers and memory. Some operations can
          be more concisely expressed with the debugging mini-language. Type
          "help assign" for details.

          Examples:
            execute ASR(R5, R4)
            execute SET(R1, 20)  SET(R2, 22)  ADD(R3, R2, R1)
        """
        if not argstr.strip():
            print("execute takes one argument.")
            return

        # Make sure there are no disallowed ops.
        for op in parse(argstr)[0]:
            if isinstance(op, Branch):
                print("execute cannot take branching operations.")
                return
            elif isinstance(op, DataOperation):
                print("execute cannot take data statements.")
                return
            elif isinstance(op, LABEL):
                print("execute cannot take labels.")
                return

        try:
            program = load_program(argstr, self.settings)
        except SystemExit:
            return

        vm = self.debugger.vm
        opc = vm.pc
        for op in program.code:
            op.execute(vm)
        vm.pc = opc

    @mutates
    def handle_goto(self, args: "List[str]") -> None:
        """
        goto <loc>
          Jump to the given location (either a line number or a label) without
          executing any of the intermediate instructions.
        """
        if len(args) != 1:
            print("goto takes one argument.")
            return

        try:
            new_pc = self.debugger.location_to_instruction_number(args[0])
        except ValueError as e:
            print("Error:", str(e))
            return
        else:
            self.debugger.vm.pc = new_pc

        self.print_current_op()

    def handle_help(self, args: "List[str]") -> None:
        """
        help
          Print a summary of all debugging commands.

        help <cmd>...
          Print a detailed help message for each command list.
        """
        if not args:
            print(HELP)
        else:
            for i, arg in enumerate(args):
                try:
                    fullarg = self.expand_command(arg)
                except HERAError:
                    print("{} is not a recognized command.".format(arg))
                else:
                    if fullarg == "quit":
                        print("quit\n  Exit the debugger.")
                    else:
                        doc = getattr(self, "handle_" + fullarg).__doc__
                        print(textwrap.dedent(doc).strip())

                if i != len(args) - 1:
                    print()

    def handle_info(self, args: "List[str]") -> None:
        """
        info <arg>...
          Print information about the current state of the program. Valid arguments to
          info are "registers", "stack", "flags" and "symbols". Arguments may be
          abbreviated with a unique prefix. The argument list defaults to "registers",
          "flags", and "stack" if not provided.
        """
        if args:
            try:
                fullargs = [self.expand_info_arg(arg) for arg in args]
            except HERAError as e:
                print("Error: " + str(e) + ".")
                return
        else:
            fullargs = ["registers", "flags", "stack"]

        for i, fullarg in enumerate(fullargs):
            if fullarg == "stack":
                self.info_stack()
            elif fullarg == "symbols":
                self.info_symbols()
            elif fullarg == "registers":
                self.info_registers()
            elif fullarg == "flags":
                self.info_flags()
            else:
                raise RuntimeError("this should never happen!")

            if i != len(fullargs) - 1:
                print()

    def handle_list(self, args: "List[str]") -> None:
        """
        list
          Print the current line of source and the three previous and next lines.

        list <n>
          Print the current line of source code and the `n` previous and next lines.
        """
        if len(args) > 1:
            print("list takes zero or one arguments.")
            return

        try:
            context = int(args[0], base=0) if args else 3
        except ValueError:
            print("Could not parse argument to list.")
            return

        if not self.debugger.finished():
            loc = self.debugger.op().loc
            self.print_range_of_ops(loc, context=context)
        else:
            print("Program has finished executing.")

    def handle_ll(self, args: "List[str]") -> None:
        """
        ll
          Print every line of the current file's source code.
        """
        if len(args) != 0:
            print("ll takes no arguments.")
            return

        if not self.debugger.finished():
            loc = self.debugger.op().loc
            self.print_range_of_ops(loc)
        else:
            print("Program has finished executing.")

    @mutates
    def handle_next(self, args: "List[str]") -> None:
        """
        next
          Execute the current line. If the current line is a CALL instruction, the
          debugger executes the entire function (including nested and recursive calls)
          and moves on to the next line. If you wish to neter over the function call,
          use `step` instead.

        next <n>
          Execute the next n instructions. This command will follow branches, so be
          careful!
        """
        if len(args) > 1:
            print("next takes zero or one arguments.")
            return

        if not self.debugger.finished():
            try:
                n = int(args[0]) if args else 1
            except ValueError:
                print("Could not parse argument to next.")
                return

            for _ in range(n):
                if self.debugger.finished():
                    break
                self.debugger.next(step=False)

        self.print_current_op()

    @mutates
    def handle_off(self, args: "List[str]") -> None:
        """
        off <f1> <f2>...
          Turn off all the HERA machine flags listed. Flags may be given in long
          form (carry-block, carry, overflow, sign, zero) or short form (cb, c, v,
          s, z).
        """
        if len(args) == 0:
            print("off takes one or more arguments.")
            return

        try:
            flags = [expand_flag(arg) for arg in args]
        except HERAError as e:
            print(e)
            return

        for flag in flags:
            setattr(self.debugger.vm, flag, False)

    @mutates
    def handle_on(self, args: "List[str]") -> None:
        """
        on <f1> <f2>...
          Turn on all the HERA machine flags listed. Flags may be given in long form
          (carry-block, carry, overflow, sign, zero) or short form (cb, c, v, s, z).
        """
        if len(args) == 0:
            print("on takes one or more arguments.")
            return

        try:
            flags = [expand_flag(arg) for arg in args]
        except HERAError as e:
            print(e)
            return

        for flag in flags:
            setattr(self.debugger.vm, flag, True)

    def handle_print(self, argstr: str) -> None:
        """
        print <x> <y> <z>...
          Print the values of all the supplied arguments. The first argument may
          optionally be a format specifier, e.g. ":xds". Each character of the string
          identifies a format in which to print the value. The following formats are
          recognized: d for decimal, x for hexadecimal, o for octal, b for binary, c
          for character literals, s for signed integers, and l for source code
          locations. When not provided, the format specifier defaults to ":xdsc".

          Examples:
            A register:        print R7
            A memory location: print @1000
            A symbol:          print some_label
            Multiple values:   print R1, R2, R3
            Formatted:         print :bl PC_ret
            Arithmetic:        print @(@(FP+1)) * 7
        """
        if not argstr:
            print("print takes one or more arguments.")
            return

        try:
            tree = miniparser.parse(argstr)
        except SyntaxError as e:
            msg = str(e)
            if msg:
                print("Parse error: {}.".format(msg))
            else:
                print("Parse error.".format(msg))
        else:
            spec = tree.fmt
            for c in spec:
                if c not in "dxobcsl":
                    print("Unknown format specifier `{}`.".format(c))
                    return

                # 'c' and 's' do not always generate output, if the given value is not a
                # char or signed integer, respectively. Output can be forced with the
                # 'C' and 'S', which we do if the user explicitly provided these
                # formats.
                spec = spec.replace("c", "C")
                spec = spec.replace("s", "S")

            try:
                if len(tree.seq) > 1:
                    for arg in tree.seq:
                        self.print_one_expr(arg, spec, with_lhs=True)
                else:
                    self.print_one_expr(tree.seq[0], spec)
            except HERAError as e:
                print("Eval error: {}.".format(e))

    def print_one_expr(self, tree, spec: str, *, with_lhs=False) -> None:
        """Print a single expression with the given format specification."""

        # Customize the format specifier depending on the type of expression.
        if isinstance(tree, RegisterNode):
            # R13 is used to hold the return value of the PC in function calls, so
            # printing the location is useful.
            if tree.value == 13 and not spec:
                spec = augment_spec(spec, "l")
        elif isinstance(tree, SymbolNode):
            if tree.value.lower() == "pc":
                spec = augment_spec(spec, "l")
            else:
                try:
                    value = self.debugger.symbol_table[tree.value]
                except KeyError:
                    raise HERAError("{} is not defined".format(tree.value))
                else:
                    if isinstance(value, Label):
                        spec = augment_spec(spec, "l")
        elif isinstance(tree, IntNode):
            if not spec:
                spec = "d"

        value = self.evaluate_node(tree)
        if with_lhs:
            print("{} = {}".format(tree, self.format_int(value, spec)))
        else:
            print(self.format_int(value, spec))

    @mutates
    def handle_restart(self, args: "List[str]") -> None:
        """
        restart
          Restart execution of the program from the beginning. All registers and
          memory cells are reset.
        """
        if len(args) != 0:
            print("restart takes no arguments.")
            return

        self.debugger.reset()
        self.print_current_op()

    @mutates
    def handle_step(self, args: "List[str]") -> None:
        """
        step
          Step into the execution of a function.  The step command is only valid when
          the current instruction is CALL.
        """
        if len(args) > 0:
            print("step takes no arguments.")
            return

        if self.debugger.op().name != "CALL":
            print("step is only valid when the current instruction is CALL.")
            return

        self.debugger.next(step=True)
        self.print_current_op()

    def handle_undo(self, args: "List[str]") -> None:
        """
        undo
          Undo the last operation that changed the state of the debugger.
        """
        if len(args) > 0:
            print("undo takes no arguments.")
            return

        if self.debugger.old is None:
            print("Nothing to undo.")
            return
        else:
            print("Undid {}.".format(self.command_history.pop()))

        self.debugger = self.debugger.old

    def expand_info_arg(self, arg: str) -> str:
        """
        Expand an abbreviated argument to info into its full name, or raise HERAError if
        the abbreviation is not recognized.
        """
        arg = arg.lower()
        if "stack".startswith(arg):
            # "stack" comes before "symbols" because "s" should resolve to "stack".
            return "stack"
        elif "symbols".startswith(arg):
            return "symbols"
        elif "registers".startswith(arg):
            return "registers"
        elif "flags".startswith(arg):
            return "flags"
        else:
            raise HERAError("unrecognized argument `{}`".format(arg))

    def info_registers(self) -> None:
        nonzero = 0
        for i, r in enumerate(self.debugger.vm.registers[1:], start=1):
            if r != 0:
                nonzero += 1
                end = ", " if i != 15 or nonzero < 15 else ""
                print("R{} = {}".format(i, r), end=end)

        if nonzero == 0:
            print("All registers set to zero.")
        elif nonzero != 15:
            print("all other registers set to zero.")
        else:
            print()

    def info_flags(self) -> None:
        vm = self.debugger.vm
        flags = []
        if vm.flag_carry_block:
            flags.append("carry-block flag is on")
        if vm.flag_carry:
            flags.append("carry flag is on")
        if vm.flag_overflow:
            flags.append("overflow flag is on")
        if vm.flag_zero:
            flags.append("zero flag is on")
        if vm.flag_sign:
            flags.append("sign flag is on")

        if len(flags) == 5:
            print("All flags are on.")
        elif len(flags) == 0:
            print("All flags are off.")
        else:
            flagstr = ", ".join(flags)
            flagstr = flagstr[0].upper() + flagstr[1:]
            print(flagstr + ", all other flags are off.")

    def info_stack(self) -> None:
        vm = self.debugger.vm
        if vm.expected_returns:
            print("Call stack (last call at bottom)")
            for call_address, return_address in vm.expected_returns:
                fname = self.debugger.find_label(call_address)
                floc = self.debugger.instruction_number_to_location(
                    call_address, append_label=False
                )
                rloc = self.debugger.instruction_number_to_location(
                    return_address - 1, append_label=False
                )
                if fname is not None:
                    print("  {} ({}, called from {})".format(fname, floc, rloc))
                else:
                    print("  {} (called from {})".format(floc, rloc))
        else:
            print("The call stack is empty.")

    def info_symbols(self) -> None:
        constants = []
        labels = []
        dlabels = []
        for key, val in self.debugger.symbol_table.items():
            if isinstance(val, Label):
                debug_info = self.debugger.program.debug_info
                labels.append("{} ({})".format(key, debug_info.labels[key]))
            elif isinstance(val, DataLabel):
                dlabels.append("{} (0x{:x})".format(key, val))
            else:
                constants.append("{} ({})".format(key, val))

        if constants:
            print("Constants: " + ", ".join(constants))

        if labels:
            print("Labels: " + ", ".join(labels))

        if dlabels:
            print("Data labels: " + ", ".join(dlabels))

    def evaluate_node(self, node: AbstractNode) -> int:
        """
        Evaluate the AST node (returned by the `miniparser` module) into an integer
        value

        A HERAError is raised on expected failures (e.g., undefined symbols) and a
        RuntimeError is raised on unexpected failures (e.g., an unknown operator) which
        generally indicate a bug elsewhere in the code.
        """
        vm = self.debugger.vm
        if isinstance(node, IntNode):
            if node.value >= 2 ** 16 or node.value < -(2 ** 15):
                raise HERAError("integer literal exceeds 16 bits")
            return node.value
        elif isinstance(node, RegisterNode):
            return vm.load_register(node.value)
        elif isinstance(node, MemoryNode):
            address = self.evaluate_node(node.address)
            return vm.load_memory(address)
        elif isinstance(node, SymbolNode):
            if node.value.lower() == "pc":
                return vm.pc
            else:
                try:
                    return self.debugger.symbol_table[node.value]
                except KeyError:
                    raise HERAError("{} is not defined".format(node.value))
        elif isinstance(node, PrefixNode):
            arg = self.evaluate_node(node.arg)
            if node.op == "-":
                result = -arg
            else:
                raise RuntimeError("unhandled prefix operator: " + node.op)

            if out_of_range(result):
                raise HERAError("overflow from unary " + node.op)

            return result
        elif isinstance(node, InfixNode):
            left = self.evaluate_node(node.left)
            right = self.evaluate_node(node.right)
            if node.op == "+":
                result = left + right
            elif node.op == "-":
                result = left - right
            elif node.op == "*":
                result = left * right
            elif node.op == "/":
                if right == 0:
                    raise HERAError("division by zero")
                result = left // right
            else:
                raise RuntimeError("unhandled infix operator: " + node.op)

            if out_of_range(result):
                raise HERAError("overflow from " + node.op)

            return result
        else:
            raise RuntimeError("unknown node type {}".format(node.__class__.__name__))

    def print_current_op(self) -> None:
        """
        Print the next operation to be executed. If the program has finished executed,
        nothing is printed.
        """
        if not self.debugger.finished():
            loc = self.debugger.op().loc
            self.print_range_of_ops(loc, context=1)
        else:
            print("Program has finished executing.")

    def print_range_of_ops(
        self, loc: Location, context: "Optional[int]" = None
    ) -> None:
        """
        Print the line indicated by `loc`, as well as `context` previous and following
        lines. If `context` is None, the whole file is printed.
        """
        lineno = loc.line - 1
        lines = loc.file_lines
        max_lineno_width = len(str(len(lines)))

        if context is None:
            lo = 0
            hi = len(lines)
        else:
            lo = max(lineno - context, 0)
            hi = min(lineno + context + 1, len(lines))

        print("[{}]\n".format(loc.path))
        for i in range(lo, hi):
            prefix = "->  " if i == lineno else "    "
            print(prefix, end="")
            print(pad(str(i + 1), max_lineno_width), end="")

            line = lines[i].rstrip()
            if line:
                print("  " + line)
            else:
                print()

    def format_int(self, v: int, spec: str) -> str:
        if not spec:
            spec = DEFAULT_SPEC

        if "l" in spec:
            spec = spec.replace("l", "")
            loc = True
        else:
            loc = False

        if loc:
            try:
                label = self.debugger.instruction_number_to_location(
                    v, append_label=False
                )
            except IndexError:
                return format_int(v, spec=spec)
            else:
                return format_int(v, spec=spec) + " [" + label + "]"
        else:
            return format_int(v, spec=spec)

    def asm_settings(self, *, code: bool, data: bool) -> Settings:
        """Override some defaults in self.settings for the assembler."""
        settings = copy.copy(self.settings)
        settings.stdout = True
        settings.code = code
        settings.data = data
        return settings


DEFAULT_SPEC = "dsc"


def augment_spec(spec: str, f: str) -> str:
    """Augment the format specifier with the additional format character."""
    if not spec:
        return augment_spec(DEFAULT_SPEC, f)
    else:
        return spec + f if f not in spec else spec


FLAG_SHORT_TO_LONG = {
    "cb": "carry_block",
    "c": "carry",
    "v": "overflow",
    "s": "sign",
    "z": "zero",
}


def expand_flag(flag: str) -> str:
    """
    Expand an abbreviated flag name into its full name, or raise HERAError if the
    abbreviation is not recognized.
    """
    flag = flag.replace("-", "_")
    if flag not in FLAG_SHORT_TO_LONG.values():
        try:
            longflag = FLAG_SHORT_TO_LONG[flag]
        except KeyError:
            raise HERAError("Unrecognized flag: `{}`.".format(flag))
        else:
            return "flag_" + longflag
    else:
        return "flag_" + flag


HELP = """\
Available commands:
    asm <op>        Show the binary machine code that the HERA operation
                    assembles to.

    assign <x> <y>  Assign the value of y to x.

    break <loc>     Set a breakpoint at the given location. When no arguments
                    are given, all current breakpoints are printed.

    clear <loc>     Clear a breakpoint at the given location.

    continue        Execute the program until a breakpoint is encountered or
                    the program terminates.

    dis <n>         Disassemble the 16-bit integer into a HERA operation.

    execute <op>    Execute a HERA operation.

    goto <loc>      Jump to the given location.

    help            Print this help message.

    info            Print information about the current state of the program.

    list <n>        Print the current lines of source code and the n previous
                    and next lines. If not provided, n defaults to 3.

    ll              Print the entire program.

    next            Execute the current line.

    off <flag>      Turn the given machine flag off.

    on <flag>       Turn the given machine flag on.

    print <x>       Print the value of x.

    restart         Restart the execution of the program from the beginning.

    step            Step into the execution of a function.

    undo            Undo the last operation.

    quit            Exit the debugger.

    <x> = <y>       Alias for "assign <x> <y>".

Commands can be abbreviated with a unique prefix, e.g. "n" for "next".

Enter "help <command>" for detailed help on a specific command."""



from hera.data import Token
from hera.lexer import Lexer


def lex_helper(text):
    return Lexer(text)


def eq(tkn1, tkn2):
    return tkn1 == tkn2 and tkn1.type == tkn2.type


def test_lexer_with_register():
    lexer = lex_helper("r1")

    assert eq(lexer.tkn, Token(Token.REGISTER, "r1"))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))


def test_lexer_with_integer():
    lexer = lex_helper("0")

    assert eq(lexer.tkn, Token(Token.INT, "0"))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))


def test_lexer_with_negative_integer():
    lexer = lex_helper("-1")

    assert eq(lexer.tkn, Token(Token.MINUS, "-"))
    assert eq(lexer.next_token(), Token(Token.INT, "1"))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))


def test_lexer_with_negative_hex_number():
    lexer = lex_helper("-0xabc")

    assert eq(lexer.tkn, Token(Token.MINUS, "-"))
    assert eq(lexer.next_token(), Token(Token.INT, "0xabc"))


def test_lexer_with_invalid_hex_number():
    lexer = lex_helper("0xghi")

    assert eq(lexer.tkn, Token(Token.INT, "0xghi"))


def test_lexer_with_invalid_octal_number():
    lexer = lex_helper("0o999")

    assert eq(lexer.tkn, Token(Token.INT, "0o999"))


def test_lexer_with_character_literal():
    lexer = lex_helper("'a'")

    assert eq(lexer.tkn, Token(Token.CHAR, "a"))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))


def test_lexer_with_consecutive_character_literals():
    lexer = lex_helper("'a''b'")

    assert eq(lexer.tkn, Token(Token.CHAR, "a"))
    assert eq(lexer.next_token(), Token(Token.CHAR, "b"))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))


def test_lexer_with_character_literal_backslash_escape():
    lexer = lex_helper("'\\n'")

    assert eq(lexer.tkn, Token(Token.CHAR, "\n"))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))


def test_lexer_with_hex_escapes():
    lexer = lex_helper("'\\x41' \"\\x41\"")

    assert eq(lexer.tkn, Token(Token.CHAR, "A"))
    assert eq(lexer.next_token(), Token(Token.STRING, "A"))


def test_lexer_with_octal_escapes():
    lexer = lex_helper("'\\0''\\12' \"\\141\"")

    assert eq(lexer.tkn, Token(Token.CHAR, "\x00"))
    assert eq(lexer.next_token(), Token(Token.CHAR, "\n"))
    assert eq(lexer.next_token(), Token(Token.STRING, "a"))


def test_lexer_with_invalid_hex_escape():
    lexer = lex_helper("'\\xa'  \"\\xgh\"")

    assert len(lexer.messages.warnings) == 1
    assert eq(lexer.tkn, Token(Token.ERROR, "over-long character literal"))
    assert eq(lexer.next_token(), Token(Token.STRING, "xgh"))
    assert len(lexer.messages.warnings) == 2


def test_lexer_with_over_long_character_literal():
    lexer = lex_helper("'abc' 10")

    assert eq(lexer.tkn, Token(Token.ERROR, "over-long character literal"))
    assert eq(lexer.next_token(), Token(Token.INT, "10"))


def test_lexer_with_string():
    lexer = lex_helper(
        """\
"a double quote: \\", a backslash: \\\\"
    """
    )

    assert eq(lexer.tkn, Token(Token.STRING, 'a double quote: ", a backslash: \\'))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))


def test_lexer_with_empty_string():
    lexer = lex_helper('""')

    assert eq(lexer.tkn, Token(Token.STRING, ""))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))


def test_lexer_with_unclosed_string_literal():
    lexer = lex_helper('"hello')

    assert eq(lexer.tkn, Token(Token.ERROR, "unclosed string literal"))


def test_lexer_with_unclosed_character_literal():
    lexer = lex_helper("'a")

    assert eq(lexer.tkn, Token(Token.ERROR, "unclosed character literal"))


def test_lexer_with_unclosed_bracketed_expression():
    lexer = lex_helper("<abc")

    assert eq(lexer.tkn, Token(Token.ERROR, "unclosed bracketed expression"))


def test_lexer_with_include():
    lexer = lex_helper('#include <HERA.h> #include "lib.hera"')

    assert eq(lexer.tkn, Token(Token.INCLUDE, "#include"))
    assert eq(lexer.next_token(), Token(Token.BRACKETED, "HERA.h"))
    assert eq(lexer.next_token(), Token(Token.INCLUDE, "#include"))
    assert eq(lexer.next_token(), Token(Token.STRING, "lib.hera"))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))


def test_lexer_with_braces():
    lexer = lex_helper("{}")

    assert eq(lexer.tkn, Token(Token.LBRACE, "{"))
    assert eq(lexer.next_token(), Token(Token.RBRACE, "}"))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))


def test_lexer_with_big_example():
    # This isn't a syntactically valid expression, but it doesn't matter to the lexer.
    lexer = lex_helper("@FP_alt R15 0xabc some_symbol :xdc -10; ,, ()+*/?")

    assert eq(lexer.tkn, Token(Token.AT, "@"))
    assert eq(lexer.next_token(), Token(Token.REGISTER, "FP_alt"))
    assert eq(lexer.next_token(), Token(Token.REGISTER, "R15"))
    assert eq(lexer.next_token(), Token(Token.INT, "0xabc"))
    assert eq(lexer.next_token(), Token(Token.SYMBOL, "some_symbol"))
    assert eq(lexer.next_token(), Token(Token.FMT, "xdc"))
    assert eq(lexer.next_token(), Token(Token.MINUS, "-"))
    assert eq(lexer.next_token(), Token(Token.INT, "10"))
    assert eq(lexer.next_token(), Token(Token.SEMICOLON, ";"))
    assert eq(lexer.next_token(), Token(Token.COMMA, ","))
    assert eq(lexer.next_token(), Token(Token.COMMA, ","))
    assert eq(lexer.next_token(), Token(Token.LPAREN, "("))
    assert eq(lexer.next_token(), Token(Token.RPAREN, ")"))
    assert eq(lexer.next_token(), Token(Token.PLUS, "+"))
    assert eq(lexer.next_token(), Token(Token.ASTERISK, "*"))
    assert eq(lexer.next_token(), Token(Token.SLASH, "/"))
    assert eq(lexer.next_token(), Token(Token.UNKNOWN, "?"))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))


def test_lexer_with_symbols_starting_with_M():
    lexer = lex_helper("more m")

    assert eq(lexer.tkn, Token(Token.SYMBOL, "more"))
    assert eq(lexer.next_token(), Token(Token.SYMBOL, "m"))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))


def test_lexer_with_single_line_comment():
    lexer = lex_helper("1 // a comment\n 2")

    assert eq(lexer.tkn, Token(Token.INT, "1"))
    assert eq(lexer.next_token(), Token(Token.INT, "2"))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))


def test_lexer_with_empty_single_line_comment():
    lexer = lex_helper("//")

    assert eq(lexer.tkn, Token(Token.EOF, ""))


def test_lexer_with_multiple_single_line_comments():
    lexer = lex_helper(
        """\
1 // one
2 // two
// no three
4 // four"""
    )

    assert eq(lexer.tkn, Token(Token.INT, "1"))
    assert eq(lexer.next_token(), Token(Token.INT, "2"))
    assert eq(lexer.next_token(), Token(Token.INT, "4"))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))


def test_lexer_with_multiline_comment():
    lexer = lex_helper(
        """\
1 /*
a multiline comment
*/ 2"""
    )

    assert eq(lexer.tkn, Token(Token.INT, "1"))
    assert eq(lexer.next_token(), Token(Token.INT, "2"))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))


def test_lexer_with_multiline_comment_abutting_value():
    lexer = lex_helper("1/*\n\n*/2")

    assert eq(lexer.tkn, Token(Token.INT, "1"))
    assert eq(lexer.next_token(), Token(Token.INT, "2"))
    assert eq(lexer.next_token(), Token(Token.EOF, ""))


def test_lexer_with_tricky_multiline_comment():
    lexer = lex_helper("/*/*** 123/ */")

    assert eq(lexer.tkn, Token(Token.EOF, ""))



import pytest
from .utils import helper

from hera.utils import to_u16
from hera.vm import VirtualMachine


@pytest.fixture
def vm():
    return VirtualMachine()


def test_LOAD_from_small_address(vm):
    vm.memory[3] = 42
    vm.registers[2] = 3

    helper(vm, "LOAD(R1, 0, R2)")

    assert vm.registers[1] == 42


def test_LOAD_from_uninitialized_address(vm):
    vm.registers[2] = 5

    helper(vm, "LOAD(R1, 0, R2)")

    assert vm.registers[1] == 0


def test_LOAD_from_large_uninitialized_address(vm):
    vm.registers[2] = 14000

    helper(vm, "LOAD(R1, 0, R2)")

    assert vm.registers[1] == 0


def test_LOAD_from_address_zero(vm):
    vm.memory[0] = 42

    helper(vm, "LOAD(R1, 0, R0)")

    assert vm.registers[1] == 42


def test_LOAD_with_offset(vm):
    vm.memory[7] = 42
    vm.registers[2] = 4

    helper(vm, "LOAD(R1, 3, R2)")

    assert vm.registers[1] == 42


def test_LOAD_sets_zero_flag(vm):
    helper(vm, "LOAD(R1, 0, R2)")

    assert vm.registers[1] == 0
    assert vm.flag_zero
    assert not vm.flag_sign


def test_LOAD_sets_sign_flag(vm):
    vm.memory[4] = to_u16(-2)
    vm.registers[2] = 4

    helper(vm, "LOAD(R1, 0, R2)")

    assert vm.registers[1] == to_u16(-2)
    assert not vm.flag_zero
    assert vm.flag_sign


def test_LOAD_ignores_other_flags(vm):
    vm.flag_carry = True
    vm.flag_overflow = True
    vm.flag_carry_block = True

    helper(vm, "LOAD(R1, 0, R2)")

    assert vm.flag_carry
    assert vm.flag_overflow
    assert vm.flag_carry_block


def test_LOAD_does_not_affect_R0(vm):
    vm.memory[2] = 5
    vm.registers[2] = 2

    helper(vm, "LOAD(R1, 0, R2)")

    assert vm.registers[0] == 0


def test_LOAD_increments_pc(vm):
    helper(vm, "LOAD(R1, 0, R2)")

    assert vm.pc == 1


def test_STORE_to_small_address(vm):
    vm.registers[1] = 42
    vm.registers[2] = 3

    helper(vm, "STORE(R1, 0, R2)")

    assert vm.memory[3] == 42


def test_STORE_to_large_address(vm):
    vm.registers[1] = 42
    vm.registers[2] = 5000

    helper(vm, "STORE(R1, 0, R2)")

    assert vm.memory[5000] == 42


def test_STORE_to_max_address(vm):
    vm.registers[1] = 42
    vm.registers[2] = (2 ** 16) - 1

    helper(vm, "STORE(R1, 0, R2)")

    assert vm.memory[(2 ** 16) - 1] == 42


def test_STORE_to_address_zero(vm):
    vm.registers[1] = 42

    helper(vm, "STORE(R1, 0, R0)")

    assert vm.memory[0] == 42


def test_STORE_with_offset(vm):
    vm.registers[1] = 42
    vm.registers[2] = 4

    helper(vm, "STORE(R1, 3, R2)")

    assert vm.memory[7] == 42


def test_STORE_ignores_all_flags(vm):
    vm.flag_zero = True
    vm.flag_sign = True
    vm.flag_carry = True
    vm.flag_overflow = True
    vm.flag_carry_block = True
    vm.registers[1] = 42

    helper(vm, "STORE(R1, 0, R2)")

    assert vm.flag_zero
    assert vm.flag_sign
    assert vm.flag_carry
    assert vm.flag_overflow
    assert vm.flag_carry_block


def test_STORE_increments_pc(vm):
    helper(vm, "STORE(R1, 0, R2)")

    assert vm.pc == 1
